from __future__ import annotations

import os
import typing as t
import tempfile
from pathlib import Path
from functools import partial

import dill

from starwhale.utils import console
from starwhale.consts import PythonRunEnv, ENV_LOG_VERBOSE_COUNT
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.utils.venv import (
    guess_python_env_mode,
    check_valid_venv_prefix,
    check_valid_conda_prefix,
)
from starwhale.utils.error import NotFoundError, NoSupportError
from starwhale.utils.process import check_call

from .model import StandaloneRuntime


class Process:
    EnvInActivatedProcess = "SW_RUNTIME_ACTIVATED_PROCESS"

    def __init__(
        self,
        prefix_path: t.Union[Path, str],
        target: t.Callable,
        args: t.Tuple = (),
        kwargs: t.Dict[str, t.Any] = {},
    ) -> None:
        self._prefix_path = Path(prefix_path).resolve()
        self._target = target
        self._args = args
        self._kwargs = kwargs
        self._mode = guess_python_env_mode(self._prefix_path)

    def __str__(self) -> str:
        return f"process: {self._target} in prefix path {self._prefix_path}"

    def __repr__(self) -> str:
        return f"process: {self._target} with args:{self._args}, kwargs:{self._kwargs} in runtime {self._prefix_path}"

    def run(self) -> None:
        partial_target = partial(self._target, *self._args, **self._kwargs)
        _, _pkl_path = tempfile.mkstemp(
            prefix="starwhale-runtime-process-", suffix=".pkl"
        )
        with open(_pkl_path, "wb") as f:
            dill.dump(partial_target, f)

        if not os.path.exists(_pkl_path):
            raise NotFoundError(f"dill file: {_pkl_path}")

        verbose = int(os.environ.get(ENV_LOG_VERBOSE_COUNT, "0"))
        try:
            cmd = [
                self._get_activate_cmd(),
                f'{self._prefix_path}/bin/python3 -c \'from starwhale.utils.debug import init_logger; init_logger({verbose}); import dill; dill.load(open("{_pkl_path}", "rb"))()\'',
            ]
            console.print(
                f":rooster: run process in the python isolated environment(prefix: {self._prefix_path})"
            )
            check_call(
                ["bash", "-c", " && ".join(cmd)],
                env={self.EnvInActivatedProcess: "1"},
                log=print,
            )
        finally:
            os.unlink(_pkl_path)

    def _get_activate_cmd(self) -> str:
        # TODO: support windows platform
        if self._mode == PythonRunEnv.VENV:
            _ap = self._prefix_path / "bin/activate"
            return f"source {_ap.absolute()}"
        elif self._mode == PythonRunEnv.CONDA:
            return f"source activate {self._prefix_path.absolute()}"
        else:
            raise NoSupportError(f"get activate command for mode: {self._mode}")

    @classmethod
    def from_runtime_uri(
        cls,
        uri: t.Union[str, URI],
        target: t.Callable,
        args: t.Tuple = (),
        kwargs: t.Dict[str, t.Any] = {},
        force_restore: bool = False,
    ) -> Process:
        _uri: URI
        if isinstance(uri, str):
            _uri = URI(uri, expected_type=URIType.RUNTIME)
        else:
            _uri = uri
        # TODO: support cloud runtime uri
        if _uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError("run process with cloud instance uri")

        runtime = StandaloneRuntime(_uri)
        venv_prefix = runtime.store.export_dir / PythonRunEnv.VENV
        conda_prefix = runtime.store.export_dir / PythonRunEnv.CONDA
        has_restored_runtime = check_valid_venv_prefix(
            venv_prefix
        ) or check_valid_conda_prefix(conda_prefix)

        if force_restore or not has_restored_runtime:
            console.print(f":snail: start to restore runtime: {uri}")
            if not runtime.store.manifest_path.exists():
                runtime.extract(force=True)
            StandaloneRuntime.restore(runtime.store.snapshot_workdir, verbose=False)

        if venv_prefix.exists():
            prefix = venv_prefix
        elif conda_prefix.exists():
            prefix = conda_prefix
        else:
            raise NotFoundError(f"{runtime.store.export_dir} cannot find valid env dir")

        return cls(
            prefix_path=prefix,
            target=target,
            args=args,
            kwargs=kwargs,
        )



import typing as t
from pathlib import Path

from starwhale.consts import SW_AUTO_DIRNAME, VERSION_PREFIX_CNT, DEFAULT_MANIFEST_NAME
from starwhale.base.type import URIType, BundleType
from starwhale.base.store import BaseStorage


class ModelStorage(BaseStorage):
    def _guess(self) -> t.Tuple[Path, str]:
        return self._guess_for_bundle()

    @property
    def bundle_type(self) -> str:
        return BundleType.MODEL

    @property
    def uri_type(self) -> str:
        return URIType.MODEL

    @property
    def src_dir_name(self) -> str:
        return "src"

    @property
    def src_dir(self) -> Path:
        return self.snapshot_workdir / self.src_dir_name

    @property
    def hidden_sw_dir(self) -> Path:
        return self.src_dir / SW_AUTO_DIRNAME

    @property
    def recover_loc(self) -> Path:
        return self._get_recover_loc_for_bundle()

    @property
    def manifest_path(self) -> Path:
        return self.snapshot_workdir / DEFAULT_MANIFEST_NAME

    @property
    def snapshot_workdir(self) -> Path:
        if self.building:
            return self.tmp_dir
        version = self.uri.object.version
        return (
            self.project_dir
            / URIType.MODEL
            / self.uri.object.name
            / version[:VERSION_PREFIX_CNT]
            / f"{version}{BundleType.MODEL}"
        )



class FileTypeError(Exception):
    pass


class FileFormatError(Exception):
    pass


class NoSupportError(Exception):
    pass


class NotFoundError(Exception):
    pass


class SWObjNameFormatError(Exception):
    pass


class UnExpectedConfigFieldError(Exception):
    pass


class URIFormatError(Exception):
    pass


class ConfigFormatError(Exception):
    pass


class ExistedError(Exception):
    pass


class MissingFieldError(Exception):
    pass


class FormatError(Exception):
    pass


class EnvironmentError(Exception):
    pass


class PythonEnvironmentError(Exception):
    pass


class FieldTypeOrValueError(Exception):
    pass


class ParameterError(Exception):
    pass


class ExclusiveArgsError(Exception):
    pass


class InvalidObjectName(Exception):
    pass


class MissingDependencyError(Exception):
    pass



from __future__ import annotations

import os
import typing as t
from pathlib import Path

from starwhale.utils.fs import blake2b_file

PathLike = t.Union[str, Path]
OptionalPathLike = t.Optional[PathLike]


class BlakeFile:
    def __init__(self, path: PathLike, blake: t.Optional[str] = None):
        self._blake = blake or blake2b_file(path)
        self._path = Path(path)

    @property
    def hash(self) -> str:
        return self._blake

    @property
    def path(self) -> Path:
        return self._path

    @property
    def size(self) -> int:
        return self._path.stat().st_size

    def exists(self) -> bool:
        return self._path.exists()

    def link(self, to: PathLike, soft: bool = False) -> None:
        Path(to).parent.mkdir(parents=True, exist_ok=True)
        # https://github.com/python/mypy/issues/10740
        # link: t.Callable[[Path, str], None] = os.symlink if soft else os.link
        os.symlink(self._path, str(to)) if soft else os.link(self._path, str(to))



import os
import re
import time
import uuid
import base64
import random
import string
import typing as t
import platform
from pathlib import Path
from datetime import datetime
from functools import cmp_to_key
from contextlib import contextmanager

import yaml
from rich.console import Console

from starwhale.consts import (
    FMT_DATETIME,
    SW_DEV_DUMMY_VERSION,
    ENV_DISABLE_PROGRESS_BAR,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.utils.error import NoSupportError

console = Console(soft_wrap=True)
now_str: t.Callable[[], str] = (
    lambda: datetime.now().astimezone().strftime(FMT_DATETIME)
)


def timestamp_to_datatimestr(timestamp: float) -> str:
    ts = time.localtime(timestamp)
    return time.strftime(FMT_DATETIME, ts)


def gen_uniq_version() -> str:
    uuid_bytes = bytearray(uuid.uuid1().bytes)
    rand_bytes = os.urandom(17)
    # ref: https://docs.python.org/3.9/library/uuid.html
    # uuid1: 32-bit time_low, 16-bit time_mid, 16-bit time_hi_version, 8-bit clock_seq_hi_variant, 8-bit clock_seq_low, 48-bit node
    # uuid only keeps time_low, time_mid, clock_seq_hi_variant and clock_seq_low
    ver_bytes = uuid_bytes[:6] + uuid_bytes[8:10] + rand_bytes
    random.shuffle(ver_bytes)
    return base64.b32encode(ver_bytes).decode("ascii").lower().strip("=")


def random_str(cnt: int = 8) -> str:
    return "".join(random.sample(string.ascii_lowercase + string.digits, cnt))


def in_dev() -> bool:
    return not in_production()


def in_production() -> bool:
    return os.environ.get("SW_PRODUCTION", "") == "1"


def in_container() -> bool:
    return os.environ.get("SW_CONTAINER", "") == "1"


def is_windows() -> bool:
    # TODO: for windows nt?
    return platform.system() == "Windows"


def is_darwin(arm: bool = False) -> bool:
    if platform.system() != "Darwin":
        return False
    return not arm or platform.processor() == "arm"


def is_linux() -> bool:
    return platform.system() in ("Linux", "Unix")


def fmt_http_server(server: str, https: bool = False) -> str:
    server = server.strip().strip("/")
    if not server:
        raise Exception("no server addr")

    if server.startswith(("http://", "https://")):
        return server
    else:
        prefix = "https" if https else "http"
        return f"{prefix}://{server}"


_bytes_map = {
    "k": 1024,
    "kb": 1024,
    "m": 1024 * 1024,
    "mb": 1024 * 1024,
    "g": 1024 * 1024 * 1024,
    "gb": 1024 * 1024 * 1024,
}


def convert_to_bytes(s: t.Union[str, int]) -> int:
    if isinstance(s, int):
        return s

    s = s.strip().lower()
    for f in ("k", "m", "g"):
        if s.endswith((f, f"{f}b")):
            return _bytes_map[f] * int(s.split(f)[0])
    else:
        return int(s)


_bytes_progress = ("B", "KB", "MB", "GB", "TB", "PB")


def pretty_bytes(b: t.Union[int, float]) -> str:
    def _c(b: t.Union[int, float], idx: int) -> str:
        if b < 1024 or (idx + 1 == len(_bytes_progress)):
            return f"{b:.2f}{_bytes_progress[idx]}"
        else:
            return _c(b / 1024, idx + 1)

    return _c(b, 0)


_valid_name_re = re.compile("^[a-zA-Z_][a-zA-Z0-9_.-]*$")


def validate_obj_name(name: str) -> t.Tuple[bool, str]:
    if len(name) < 1 or len(name) > 80:
        return (
            False,
            f"length should be between 1 and 80, but {name} has {len(name)} characters",
        )

    if not _valid_name_re.match(name):
        return (
            False,
            f"A name MUST only consist of letters A-Z a-z, digits 0-9, the hyphen character -, the underscore character _ and the dot character ., current name:{name}",
        )

    return True, ""


def get_downloadable_sw_version() -> str:
    _v = STARWHALE_VERSION
    return "" if _v == SW_DEV_DUMMY_VERSION else _v


def get_field(item: t.Dict, field: str):  # type:ignore
    for k in field.split("."):
        item = item.get(k)  # type:ignore
    return item


def snake_to_camel(snake: str) -> str:
    parts = snake.split("_")
    return "".join(i.title() for i in parts)


class Order:
    def __init__(self, field: str, reverse: bool = False):
        self.field = field
        self.reverse = reverse


def sort_obj_list(
    data: t.Sequence, orders: t.List[Order]
) -> t.List[t.Dict[str, t.Any]]:
    def cmp(a: t.Any, b: t.Any) -> t.Any:
        return (a > b) - (a < b)

    def compare(lhs: t.Dict, rhs: t.Dict) -> int:
        m = 0
        for o in orders:
            m = cmp(get_field(lhs, o.field), get_field(rhs, o.field))
            m = o.reverse and -m or m
            if m != 0:
                return m
        return m

    return sorted(data, key=cmp_to_key(compare))


def load_yaml(path: t.Union[str, Path]) -> t.Any:
    """load_yaml loads yaml from path, it may raise exception such as yaml.YAMLError"""
    with open(path) as f:
        return yaml.safe_load(f)


def get_current_shell() -> str:
    import shellingham

    return str(shellingham.detect_shell()[0])


@contextmanager
def disable_progress_bar() -> t.Generator[None, None, None]:
    old_flag = os.environ.get(ENV_DISABLE_PROGRESS_BAR, "")

    os.environ[ENV_DISABLE_PROGRESS_BAR] = "0"
    try:
        yield
    finally:
        os.environ[ENV_DISABLE_PROGRESS_BAR] = old_flag


def make_dir_gitignore(d: Path) -> None:
    from starwhale.utils.fs import ensure_dir, ensure_file

    if d.exists() and not d.is_dir():
        raise NoSupportError(f"{d} is not dir")

    ensure_dir(d)
    ensure_file(d / ".gitignore", "*")


def load_dotenv(fpath: Path) -> None:
    if not fpath.exists():
        return

    with fpath.open("r") as f:
        for line in f.readlines():
            line = line.strip()
            if not line or line.startswith("#") or "=" not in line:
                continue

            k, v = line.split("=", 1)
            os.environ[k.strip()] = v.strip()


def pretty_merge_list(lst: t.List[int]) -> str:
    _r = []
    lst = sorted(lst)
    if not lst:
        return ""

    _start, _end = lst[0], lst[0]
    for _cur in lst[1:]:
        if _end + 1 == _cur or _end == _cur:
            _end = _cur
        else:
            _r.append(str(_start) if _start == _end else f"{_start}-{_end}")
            _start = _end = _cur

    _r.append(str(_start) if _start == _end else f"{_start}-{_end}")
    return ",".join(_r)



import typing as t
from inspect import iscoroutinefunction
from urllib.error import HTTPError

import requests
from tenacity import Retrying
from tenacity.stop import stop_after_attempt
from tenacity.retry import retry_if_exception
from tenacity._asyncio import AsyncRetrying
from requests.exceptions import HTTPError as RequestsHTTPError

# https://docs.microsoft.com/en-us/azure/architecture/best-practices/retry-service-specific#general-rest-and-retry-guidelines
_RETRY_HTTP_STATUS_CODES = (408, 429, 500, 502, 503, 504)


class retry_if_http_exception(retry_if_exception):
    def __init__(self, status_codes: t.Optional[t.Sequence[int]] = None) -> None:
        self.status_codes = status_codes or _RETRY_HTTP_STATUS_CODES

        def _predicate(e: BaseException) -> bool:
            if isinstance(e, RequestsHTTPError) and isinstance(
                e.response, requests.Response
            ):
                return e.response.status_code in self.status_codes
            elif isinstance(e, HTTPError):
                return e.code in self.status_codes
            else:
                return False

        super().__init__(_predicate)


def http_retry(*args: t.Any, **kw: t.Any) -> t.Any:

    # support http_retry and http_retry()
    if len(args) == 1 and callable(args[0]):
        return http_retry()(args[0])
    else:

        def wrap(f: t.Callable) -> t.Any:
            _attempts = kw.get("attempts", 3)
            _cls = AsyncRetrying if iscoroutinefunction(f) else Retrying
            return _cls(
                *args,
                reraise=True,
                stop=stop_after_attempt(_attempts),
                retry=retry_if_http_exception(_RETRY_HTTP_STATUS_CODES),
            ).wraps(f)

        return wrap



import os

from setuptools import setup, find_packages

install_requires = [
    "click-option-group==0.5.5",
    "click>=8.0.4",
    "shellingham>=1.4.0",
    # flake8 require importlib-metadata < 4.3.0
    # importlib-metadata supplies a backport of 'importlib.metadata'; Python 3.8 makes 'importlib' part of the standard lib.
    "importlib-metadata>=4.0.0, <=4.2.0;python_version < '3.8'",
    "attrs>=21.4.0",
    "pyyaml==6.0",
    "cattrs>=1.7.1",
    "requests>=2.1.0",
    "requests-toolbelt>=0.9.0",
    "loguru==0.6.0",
    "conda-pack==0.6.0",
    "virtualenv>=13.0.0",
    "fs>=2.4.0",
    "typing-extensions>=4.0.0",
    "commonmark>=0.9.1",
    "textual==0.1.18",
    "jsonlines==3.0.0",
    # botocore(>=1.29.14) fixed the cgi deprecation warning which has been updated in the boto3 1.26.14 version
    "boto3>=1.26.14",
    "scikit-learn>=0.20.0",
    # Python 3.11 needs dill >= 0.3.6
    "dill>=0.3.6",
    "packaging>=21.3",
    "pyarrow>=8.0.0",
    "Jinja2>=3.1.2",
    "tenacity>=8.0.1",
    "gradio~=3.15.0",
    # for system monitor
    "psutil>=5.5.0",
    "GitPython>=3.1.24",
    "fastapi>=0.71.0",
    "filelock",
]

extras_require = {
    "image": ["pillow"],
    "audio": ["soundfile"],
}

all_requires = list(
    set(install_requires + [r for v in extras_require.values() for r in v])
)
extras_require["all"] = all_requires


def _format_version() -> str:
    _v = os.environ.get("PYPI_RELEASE_VERSION", "0.0.0.dev")
    _v = _v.lstrip("v").replace("-", ".")
    _vs = _v.split(".", 3)
    if len(_vs) == 4:
        _vs[-1] = _vs[-1].replace(".", "")
        return ".".join(_vs)
    else:
        return _v


setup(
    name="starwhale",
    author="Starwhale Team",
    author_email="developer@starwhale.ai",
    version=_format_version(),
    description="An MLOps Platform for Model Evaluation",
    long_description=open("../README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    keywords="MLOps, AI, Starwhale, Model Evaluation",
    url="https://github.com/star-whale/starwhale",
    license="Apache License 2.0",
    packages=find_packages(exclude=["ez_setup", "tests*"]),
    package_data={"": ["core/runtime/template/Dockerfile.tmpl", "web/ui/**"]},
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    zip_safe=False,
    entry_points="""
      [console_scripts]
      swcli = starwhale.cli:cli
      sw = starwhale.cli:cli
      starwhale = starwhale.cli:cli
      """,
    python_requires=">=3.7, <3.12",
    scripts=[
        "scripts/sw-docker-entrypoint",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation :: CPython",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        "Topic :: System :: Logging",
        "Topic :: System :: Monitoring",
    ],
)



import os
import sys
import typing as t
from pathlib import Path

from rich import box
from rich.text import Text
from rich.panel import Panel
from rich.table import Table
from rich.console import Group

from starwhale.utils import (
    docker,
    console,
    process,
    load_yaml,
    pretty_bytes,
    in_production,
)
from starwhale.consts import (
    FileFlag,
    DefaultYAMLName,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
)
from starwhale.base.uri import URI
from starwhale.utils.fs import cmp_file_content
from starwhale.base.type import URIType, InstanceType
from starwhale.base.view import BaseTermView
from starwhale.consts.env import SWEnv
from starwhale.utils.error import FieldTypeOrValueError
from starwhale.core.model.store import ModelStorage
from starwhale.core.runtime.model import StandaloneRuntime
from starwhale.core.runtime.process import Process as RuntimeProcess

from .model import Model, StandaloneModel


class ModelTermView(BaseTermView):
    def __init__(self, model_uri: str) -> None:
        super().__init__()

        self.raw_uri = model_uri
        self.uri = URI(model_uri, expected_type=URIType.MODEL)
        self.model = Model.get_model(self.uri)

    @BaseTermView._simple_action_print
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.remove(force)

    @BaseTermView._simple_action_print
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        return self.model.recover(force)

    @BaseTermView._header
    def info(self, fullname: bool = False) -> None:
        self._print_info(self.model.info(), fullname=fullname)

    @BaseTermView._only_standalone
    def diff(self, compare_uri: URI, show_details: bool) -> None:
        r = self.model.diff(compare_uri)
        text_details: t.List[Panel] = []
        table = Table(box=box.SIMPLE, expand=False, show_lines=True)
        table.add_column("file")
        table.add_column("path")
        table.add_column(
            f"base: {self.model.version[:SHORT_VERSION_CNT]}",
            style="magenta",
            justify="left",
        )
        table.add_column(
            f"compare: {compare_uri.object.version[:SHORT_VERSION_CNT]}",
            style="cyan",
            justify="left",
        )
        b_files = r["base_version"]
        c_files = r["compare_version"]
        for _p in r["all_paths"]:
            _flag = c_files[_p].flag
            if _flag == FileFlag.ADDED:
                _flag_txt = f"[blue]{_flag}"
            elif _flag == FileFlag.DELETED:
                _flag_txt = f"[red]{_flag}"
            elif _flag == FileFlag.UPDATED:
                _flag_txt = f"[yellow]{_flag}"
                if show_details:
                    text_details.append(
                        Panel(
                            Text(
                                text="".join(
                                    cmp_file_content(
                                        base_path=b_files[_p].path,
                                        cmp_path=c_files[_p].path,
                                    )
                                )
                            ),
                            title=_p,
                        )
                    )
            else:
                _flag_txt = f"[green]{_flag}"
            table.add_row(
                os.path.basename(_p),
                _p,
                "existed" if _p in b_files else "-",
                f"[green]{_flag_txt}",
            )

        console.print(Panel(table, title="diff overview"))
        if show_details:
            console.print(Panel(Group(*text_details), title="diff details"))

    @BaseTermView._pager
    @BaseTermView._header
    def history(self, fullname: bool = False) -> t.List[t.Dict[str, t.Any]]:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        return self._print_history(
            title="Model History List", history=self.model.history(), fullname=fullname
        )

    @classmethod
    @BaseTermView._only_standalone
    def eval(
        cls,
        project: str,
        target: str,
        dataset_uris: t.List[str],
        version: str = "",
        yaml_name: str = DefaultYAMLName.MODEL,
        step: str = "",
        task_index: t.Optional[int] = None,
        task_num: int = 0,
        runtime_uri: str = "",
        use_docker: bool = False,
        gencmd: bool = False,
        image: str = "",
    ) -> None:
        if use_docker:
            if not runtime_uri and not image:
                raise FieldTypeOrValueError(
                    "runtime_uri and image both are none when use_docker"
                )
            if runtime_uri:
                runtime = StandaloneRuntime(
                    URI(runtime_uri, expected_type=URIType.RUNTIME)
                )
                image = runtime.store.get_docker_base_image()
            mnt_paths = (
                [os.path.abspath(target)]
                if in_production() or (os.path.exists(target) and os.path.isdir(target))
                else []
            )
            env_vars = {SWEnv.runtime_version: runtime_uri} if runtime_uri else {}
            cmd = docker.gen_swcli_docker_cmd(
                image,
                env_vars=env_vars,
                mnt_paths=mnt_paths,
            )
            console.rule(":elephant: docker cmd", align="left")
            console.print(f"{cmd}\n")
            if gencmd:
                return
            process.check_call(cmd, shell=True)
            return

        kw = dict(
            project=project,
            version=version,
            workdir=cls._get_workdir(target),
            dataset_uris=dataset_uris,
            step_name=step,
            task_index=task_index,
            task_num=task_num,
            model_yaml_name=yaml_name,
        )
        if not in_production() and runtime_uri:
            RuntimeProcess.from_runtime_uri(
                uri=runtime_uri,
                target=StandaloneModel.eval_user_handler,
                kwargs=kw,
            ).run()
        else:
            StandaloneModel.eval_user_handler(**kw)  # type: ignore

    @classmethod
    @BaseTermView._only_standalone
    def fine_tune(
        cls,
        project: str,
        target: str,
        dataset_uris: t.List[str],
        yaml_name: str = DefaultYAMLName.MODEL,
        runtime_uri: str = "",
        model_uri: str = "",
    ) -> None:
        if target and model_uri:
            console.print("workdir and model can not both set together")
            sys.exit(1)
        if not target and not model_uri:
            console.print("workdir or model needs to be set")
            sys.exit(1)

        if target:
            workdir = cls._get_workdir(target)
        else:
            _m = StandaloneModel(URI(model_uri, expected_type=URIType.MODEL))
            workdir = _m.store.src_dir

        kw = dict(
            project=project,
            workdir=workdir,
            dataset_uris=dataset_uris,
            model_yaml_name=yaml_name,
        )
        if not in_production() and runtime_uri:
            RuntimeProcess.from_runtime_uri(
                uri=runtime_uri,
                target=StandaloneModel.fine_tune,
                kwargs=kw,
            ).run()
        else:
            StandaloneModel.fine_tune(**kw)  # type: ignore

    @classmethod
    def _get_workdir(cls, target: str) -> Path:
        if in_production() or (os.path.exists(target) and os.path.isdir(target)):
            workdir = Path(target)
        else:
            _uri = URI(target, URIType.MODEL)
            _store = ModelStorage(_uri)
            workdir = _store.src_dir
        return workdir

    @classmethod
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or {}
        _uri = URI(project_uri, expected_type=URIType.PROJECT)
        cls.must_have_project(_uri)
        fullname = fullname or (_uri.instance_type == InstanceType.CLOUD)
        _models, _pager = Model.list(_uri, page, size, filters)
        _data = BaseTermView.list_data(_models, show_removed, fullname)
        return _data, _pager

    @classmethod
    @BaseTermView._only_standalone
    def build(
        cls,
        workdir: t.Union[str, Path],
        project: str,
        yaml_path: t.Union[str, Path],
        runtime_uri: str = "",
    ) -> URI:
        workdir = Path(workdir)
        _config = load_yaml(yaml_path)
        _model_uri = cls.prepare_build_bundle(
            project=project, bundle_name=_config.get("name"), typ=URIType.MODEL
        )
        _m = Model.get_model(_model_uri)
        kwargs = {"yaml_path": yaml_path}
        if runtime_uri:
            RuntimeProcess.from_runtime_uri(
                uri=runtime_uri,
                target=_m.build,
                args=(workdir,),
                kwargs=kwargs,
            ).run()
        else:
            _m.build(workdir, **kwargs)
        return _model_uri

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        force: bool = False,
        dest_local_project_uri: str = "",
    ) -> None:
        Model.copy(src_uri, dest_uri, force, dest_local_project_uri)
        console.print(":clap: copy done.")

    @BaseTermView._header
    def tag(
        self, tags: t.List[str], remove: bool = False, ignore_errors: bool = False
    ) -> None:
        if remove:
            console.print(f":golfer: remove tags [red]{tags}[/] @ {self.uri}...")
            self.model.remove_tags(tags, ignore_errors)
        else:
            console.print(f":surfer: add tags [red]{tags}[/] @ {self.uri}...")
            self.model.add_tags(tags, ignore_errors)

    @classmethod
    @BaseTermView._only_standalone
    def serve(
        cls,
        target: str,
        model_yaml: str,
        runtime_uri: str,
        model_uri: str,
        host: str,
        port: int,
    ) -> None:
        if target and model_uri:
            console.print("workdir and model can not both set together")
            sys.exit(1)
        if not target and not model_uri:
            console.print("workdir or model needs to be set")
            sys.exit(1)

        if target:
            workdir = cls._get_workdir(target)
        else:
            _m = StandaloneModel(URI(model_uri, expected_type=URIType.MODEL))
            workdir = _m.store.src_dir

        kw = dict(
            model_yaml=model_yaml,
            workdir=workdir,
            host=host,
            port=port,
        )

        if runtime_uri:
            RuntimeProcess.from_runtime_uri(
                uri=runtime_uri,
                target=StandaloneModel.serve,
                kwargs=kw,
            ).run()
        else:
            StandaloneModel.serve(**kw)  # type: ignore


class ModelTermViewRich(ModelTermView):
    @classmethod
    @BaseTermView._pager
    def list(
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> t.Tuple[t.List[t.Dict[str, t.Any]], t.Dict[str, t.Any]]:
        filters = filters or []
        _models, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "tags": lambda x: ",".join(x),
            "size": lambda x: pretty_bytes(x),
            "runtime": cls.place_holder_for_empty(""),
        }

        cls.print_header()
        cls.print_table("Model List", _models, custom_column=custom_column)
        return _models, _pager


class ModelTermViewJson(ModelTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        project_uri: str = "",
        fullname: bool = False,
        show_removed: bool = False,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.List[str]] = None,
    ) -> None:
        filters = filters or []
        _models, _pager = super().list(
            project_uri, fullname, show_removed, page, size, filters
        )
        cls.pretty_json(_models)

    def info(self, fullname: bool = False) -> None:
        self.pretty_json(self.get_info_data(self.model.info(), fullname))

    def history(self, fullname: bool = False) -> None:
        fullname = fullname or self.uri.instance_type == InstanceType.CLOUD
        self.pretty_json(BaseTermView.get_history_data(self.model.history(), fullname))


def get_term_view(ctx_obj: t.Dict) -> t.Type[ModelTermView]:
    return ModelTermViewJson if ctx_obj.get("output") == "json" else ModelTermViewRich



from __future__ import annotations

import os
import queue
import typing as t
import threading
from functools import total_ordering

import loguru
from loguru import logger as _logger

from starwhale.consts import HTTPMethod
from starwhale.base.uri import URI
from starwhale.base.type import URIType, InstanceType
from starwhale.base.cloud import CloudRequestMixed
from starwhale.utils.error import ParameterError
from starwhale.core.dataset.tabular import (
    TabularDataset,
    TabularDatasetRow,
    DEFAULT_CONSUMPTION_BATCH_SIZE,
    TabularDatasetSessionConsumption,
)

_DEFAULT_LOADER_CACHE_SIZE = 20

if t.TYPE_CHECKING:
    from .model import Dataset


@total_ordering
class DataRow:
    class _Features(dict):
        _PROTECTED_PREFIX = "_starwhale_"

        def __setitem__(self, key: t.Any, value: t.Any) -> None:
            super().__setitem__(key, value)
            self._patch_shadow_dataset(key, value)

        def __delitem__(self, key: t.Any) -> None:
            super().__delitem__(key)
            # datastore will ignore none column for scanning by default
            self._patch_shadow_dataset(key, None)

        def __setattr__(self, name: str, value: t.Any) -> None:
            if name.startswith(self._PROTECTED_PREFIX):
                super().__setattr__(name, value)
            else:
                self[name] = value

        def __getattr__(self, name: str) -> t.Any:
            if name.startswith(self._PROTECTED_PREFIX):
                return super().__getattribute__(name)
            elif name in self:
                return self[name]
            else:
                raise AttributeError(f"Not found attribute: {name}")

        def __delattr__(self, name: str) -> None:
            if name.startswith(self._PROTECTED_PREFIX):
                raise RuntimeError(f"cannot delete internal attribute: {name}")
            elif name in self:
                del self[name]
            else:
                raise AttributeError(f"Not found attribute: {name}")

        def _patch_shadow_dataset(self, key: t.Any, value: t.Any) -> None:
            # TODO: merge batch update
            ds = getattr(self, "_starwhale_shadow_dataset", None)
            if ds is not None:
                # pass dict type into dataset __setitem__, not the DataRow._Features type
                ds.__setitem__(self._starwhale_index, {key: value})

        def _with_shadow_dataset(
            self, dataset: Dataset, index: t.Union[str, int]
        ) -> DataRow._Features:
            from .model import Dataset

            if not isinstance(dataset, Dataset):
                raise TypeError(
                    f"shadow dataset only supports starwhale.Dataset type: {dataset}"
                )

            self._starwhale_shadow_dataset = dataset
            self._starwhale_index = index
            return self

    def __init__(
        self,
        index: t.Union[str, int],
        features: t.Dict,
        shadow_dataset: t.Optional[Dataset] = None,
    ) -> None:
        if not isinstance(index, (str, int)):
            raise TypeError(f"index({index}) is not int or str type")
        self.index = index

        if not isinstance(features, dict):
            raise TypeError(f"features({features}) is not dict type")
        self.features = DataRow._Features(features)
        if shadow_dataset is not None:
            self.features._with_shadow_dataset(shadow_dataset, index)
        self._shadow_dataset = shadow_dataset

    def _patch_shadow_dataset(self, dataset: Dataset) -> None:
        if self._shadow_dataset is not None and self._shadow_dataset is not dataset:
            raise RuntimeError("shadow dataset has already been set")

        if dataset is not None:
            self._shadow_dataset = dataset
            self.features._with_shadow_dataset(dataset, self.index)

    def __str__(self) -> str:
        return f"{self.index}"

    def __repr__(self) -> str:
        return f"index:{self.index}, features:{self.features}, shadow dataset: {self._shadow_dataset}"

    def __iter__(self) -> t.Iterator:
        return iter(self._get_items())

    def __getitem__(self, i: int) -> t.Any:
        return self._get_items()[i]

    def __len__(self) -> int:
        return len(self._get_items())

    def _get_items(self) -> t.Tuple:
        return (self.index, self.features)

    def __lt__(self, obj: DataRow) -> bool:
        return str(self.index) < str(obj.index)

    def __eq__(self, obj: t.Any) -> bool:
        return bool(self.index == obj.index and self.features == obj.features)


_TMetaQItem = t.Optional[t.Union[TabularDatasetRow, Exception]]
_TRowQItem = t.Optional[t.Union[DataRow, Exception]]


class DataLoader:
    def __init__(
        self,
        dataset_uri: URI,
        start: t.Optional[t.Any] = None,
        end: t.Optional[t.Any] = None,
        logger: t.Optional[loguru.Logger] = None,
        session_consumption: t.Optional[TabularDatasetSessionConsumption] = None,
        cache_size: int = _DEFAULT_LOADER_CACHE_SIZE,
        num_workers: int = 2,
        dataset_scan_revision: str = "",
    ):
        self.dataset_uri = dataset_uri
        self.logger = logger or _logger
        self.start = start
        self.end = end
        self.dataset_scan_revision = dataset_scan_revision

        # TODO: refactor TabularDataset with dataset_uri
        # TODO: refactor dataset, tabular_dataset and standalone dataset module
        self.tabular_dataset = TabularDataset.from_uri(
            dataset_uri,
            start=start,
            end=end,
            data_datastore_revision=self.dataset_scan_revision,
        )
        self.session_consumption = session_consumption
        self.last_processed_range: t.Optional[t.Tuple[t.Any, t.Any]] = None

        if num_workers <= 0:
            raise ValueError(
                f"num_workers({num_workers}) must be a positive int number"
            )
        self._num_workers = num_workers

        if cache_size <= 0:
            raise ValueError(f"cache_size({cache_size}) must be a positive int number")
        self._cache_size = cache_size

    def _sign_uris(self, uris: t.List[str]) -> dict:
        _batch_size = (
            self.session_consumption.batch_size
            if self.session_consumption
            else DEFAULT_CONSUMPTION_BATCH_SIZE
        )
        r = CloudRequestMixed.do_http_request(
            f"/project/{self.dataset_uri.project}/{self.dataset_uri.object.typ}/{self.dataset_uri.object.name}/uri/sign-links",
            method=HTTPMethod.POST,
            instance_uri=self.dataset_uri,
            params={
                "expTimeMillis": int(
                    os.environ.get("SW_MODEL_PROCESS_UNIT_TIME_MILLIS", "60000")
                )
                * _batch_size,
            },
            json=uris,
            use_raise=True,
        ).json()
        return r["data"]  # type: ignore

    def _iter_meta(self) -> t.Generator[TabularDatasetRow, None, None]:
        if not self.session_consumption:
            # TODO: refactor for batch-signed urls
            for row in self.tabular_dataset.scan():
                yield row
        else:
            while True:
                # TODO: tune last processed range for multithread
                pk = [self.last_processed_range] if self.last_processed_range else None
                rt = self.session_consumption.get_scan_range(pk)
                self.last_processed_range = rt
                if rt is None:
                    break

                if self.dataset_uri.instance_type == InstanceType.CLOUD:
                    for rows in self.tabular_dataset.scan_batch(
                        rt[0], rt[1], self.session_consumption.batch_size
                    ):
                        _links = []
                        for row in rows:
                            for at in row.artifacts:
                                at.owner = self.dataset_uri
                                if at.link:
                                    _links.append(at.link)
                        uri_dict = self._sign_uris([lk.uri for lk in _links])
                        for lk in _links:
                            lk.signed_uri = uri_dict.get(lk.uri, "")

                        for row in rows:
                            yield row
                else:
                    for row in self.tabular_dataset.scan(rt[0], rt[1]):
                        yield row

    def _iter_meta_with_queue(self, mq: queue.Queue[_TMetaQItem]) -> None:
        # TODO: tune last processed range
        try:
            for meta in self._iter_meta():
                if meta and isinstance(meta, TabularDatasetRow):
                    mq.put(meta)
        except Exception as e:
            mq.put(e)
            raise

        for _ in range(0, self._num_workers):
            mq.put(None)

    def _unpack_row(
        self,
        row: TabularDatasetRow,
        skip_fetch_data: bool = False,
        shadow_dataset: t.Optional[Dataset] = None,
    ) -> DataRow:
        for artifact in row.artifacts:
            artifact.owner = self.dataset_uri
            if not skip_fetch_data:
                artifact.fetch_data()
        return DataRow(
            index=row.id, features=row.features, shadow_dataset=shadow_dataset
        )

    def _unpack_row_with_queue(
        self,
        in_mq: queue.Queue[_TMetaQItem],
        out_mq: queue.Queue[_TRowQItem],
        skip_fetch_data: bool = False,
        shadow_dataset: t.Optional[Dataset] = None,
    ) -> None:
        while True:
            meta = in_mq.get(block=True, timeout=None)
            if meta is None:
                break
            elif isinstance(meta, Exception):
                out_mq.put(meta)
                raise meta
            else:
                try:
                    row = self._unpack_row(
                        meta, skip_fetch_data, shadow_dataset=shadow_dataset
                    )
                    if row and isinstance(row, DataRow):
                        out_mq.put(row)
                except Exception as e:
                    out_mq.put(e)
                    raise

        out_mq.put(None)

    def __iter__(
        self,
    ) -> t.Generator[DataRow, None, None]:
        meta_fetched_queue: queue.Queue[_TMetaQItem] = queue.Queue(4 * self._cache_size)
        row_unpacked_queue: queue.Queue[_TRowQItem] = queue.Queue(self._cache_size)

        meta_fetcher = threading.Thread(
            name="meta-fetcher",
            target=self._iter_meta_with_queue,
            args=(meta_fetched_queue,),
            daemon=True,
        )
        meta_fetcher.start()

        rows_unpackers = []
        for i in range(0, self._num_workers):
            _t = threading.Thread(
                name=f"row-unpacker-{i}",
                target=self._unpack_row_with_queue,
                args=(meta_fetched_queue, row_unpacked_queue),
                daemon=True,
            )
            _t.start()
            rows_unpackers.append(_t)

        done_unpacker_cnt = 0
        while True:
            row = row_unpacked_queue.get(block=True, timeout=None)
            if row is None:
                done_unpacker_cnt += 1
                if done_unpacker_cnt == self._num_workers:
                    break
            elif isinstance(row, Exception):
                raise row
            else:
                yield row

        self.logger.debug(
            "queue details:"
            f"meta fetcher(qsize:{meta_fetched_queue.qsize()}, alive: {meta_fetcher.is_alive()}), "
            f"row unpackers(qsize:{row_unpacked_queue.qsize()}, alive: {[t.is_alive() for t in rows_unpackers]})"
        )

    def __str__(self) -> str:
        return f"DataLoader for {self.dataset_uri}, range:[{self.start},{self.end}], use consumption:{bool(self.session_consumption)}"

    def __repr__(self) -> str:
        return (
            f"DataLoader for {self.dataset_uri}, consumption:{self.session_consumption}"
        )


def get_data_loader(
    dataset_uri: t.Union[str, URI],
    start: t.Optional[t.Any] = None,
    end: t.Optional[t.Any] = None,
    session_consumption: t.Optional[TabularDatasetSessionConsumption] = None,
    logger: t.Optional[loguru.Logger] = None,
    cache_size: int = _DEFAULT_LOADER_CACHE_SIZE,
    num_workers: int = 2,
    dataset_scan_revision: str = "",
) -> DataLoader:

    if session_consumption:
        sc_start = session_consumption.session_start  # type: ignore
        sc_end = session_consumption.session_end  # type: ignore
        if sc_start != start or sc_end != end:
            raise ParameterError(
                f"star-end range keys not match, session_consumption:[{sc_start}, {sc_end}], loader:[{start}, {end}]"
            )

    if isinstance(dataset_uri, str):
        dataset_uri = URI(dataset_uri, expected_type=URIType.DATASET)

    return DataLoader(
        dataset_uri,
        start=start,
        end=end,
        session_consumption=session_consumption,
        logger=logger or _logger,
        cache_size=cache_size,
        num_workers=num_workers,
        dataset_scan_revision=dataset_scan_revision,
    )



import typing as t
import collections

from torch.utils.data import IterableDataset

import starwhale.core.dataset.type as sw_type
from starwhale.api._impl.dataset import Dataset

__all__ = ["TorchIterableDataset", "default_transform"]


def _dummy_transform(data: t.Any) -> t.Any:
    return data


def default_transform(data: t.Any) -> t.Any:
    data_type = type(data)
    if isinstance(
        data,
        (sw_type.Audio, sw_type.Image, sw_type.GrayscaleImage, sw_type.BoundingBox),
    ):
        return data.to_tensor()
    elif isinstance(data, (sw_type.Binary, sw_type.Video)):
        return data.to_bytes()
    elif isinstance(data, sw_type.Text):
        return data.to_str()
    elif isinstance(data, collections.abc.Mapping):  # type: ignore
        try:
            return data_type({k: default_transform(v) for k, v in data.items()})
        except TypeError:
            # The mapping type may not support __init__(iterable)
            return {k: default_transform(v) for k, v in data.items()}
    elif isinstance(data, collections.abc.Sequence):  # type: ignore
        if isinstance(data, (str, bytes)):
            return data
        else:
            try:
                return data_type([default_transform(d) for d in data])
            except TypeError:
                # The sequence type may not support __init__(iterable), (e.g.: range)
                return [default_transform(d) for d in data]
    else:
        return data


class TorchIterableDataset(IterableDataset):
    def __init__(
        self,
        dataset: Dataset,
        transform: t.Optional[t.Callable] = None,
        drop_index: bool = True,
        skip_default_transform: bool = False,
    ) -> None:
        super().__init__()
        self.dataset = dataset
        if transform is not None:
            self.transform = transform
        else:
            self.transform = (
                _dummy_transform if skip_default_transform else default_transform
            )

        self.drop_index = drop_index

    def __iter__(self) -> t.Iterator:
        _t = self.transform
        for row in self.dataset:
            if self.drop_index:
                yield _t(row.features)
            else:
                yield _t(row.index), _t(row.features)

    def __len__(self) -> int:
        return len(self.dataset)

    def __str__(self) -> str:
        return f"TorchIterableDataset from Starwhale Dataset: {self.dataset}, drop_index:{self.drop_index}, transform: {self.transform}"

    __repr__ = __str__



import time
import subprocess
from typing import Any, Dict, List

import requests

from .common import (
    FileReader,
    FileWriter,
    TaskRunner,
    ChunkBuffer,
    BrokerReader,
    BrokerWriter,
    PeriodicRunner,
    check_status_code,
)


class CommandExecutor(TaskRunner):
    def __init__(self, name: str, broker_url: str, args: List[str]):
        super().__init__(name)
        self.broker_url = broker_url
        self.args = args

    def run(self) -> None:
        stdin_reader = None
        stdin_writer = None
        stdout_reader = None
        stdout_writer = None
        stderr_reader = None
        stderr_writer = None
        keep_alive = None
        try:
            keep_alive = PeriodicRunner(f"{self.name}-keep-alive", self.keep_alive, 1)
            keep_alive.start()
            with subprocess.Popen(
                self.args,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
            ) as p:
                assert p.stdin is not None
                stdin_buffer = ChunkBuffer(256 * 1024, 64)
                stdin_writer = FileWriter(
                    f"{self.name}-stdin-writer", p.stdin, stdin_buffer
                )
                stdin_writer.start()
                stdin_reader = BrokerReader(
                    f"{self.name}-stdin-reader",
                    f"{self.broker_url}/file/stdin",
                    stdin_buffer,
                )
                stdin_reader.start()
                assert p.stdout is not None
                stdout_buffer = ChunkBuffer(256 * 1024, 64)
                stdout_reader = FileReader(
                    f"{self.name}-stdout-reader", p.stdout, stdout_buffer
                )
                stdout_reader.start()
                stdout_writer = BrokerWriter(
                    f"{self.name}-stdout-writer",
                    f"{self.broker_url}/file/stdout",
                    stdout_buffer,
                )
                stdout_writer.start()
                assert p.stderr is not None
                stderr_buffer = ChunkBuffer(256 * 1024, 64)
                stderr_reader = FileReader(
                    f"{self.name}-stderr-reader", p.stderr, stderr_buffer
                )
                stderr_reader.start()
                stderr_writer = BrokerWriter(
                    f"{self.name}-stderr-writer",
                    f"{self.broker_url}/file/stderr",
                    stderr_buffer,
                )
                stderr_writer.start()
                while not self.stopped:
                    try:
                        exit_code = p.wait(0.5)
                        break
                    except subprocess.TimeoutExpired:
                        pass
                if self.stopped:
                    p.terminate()
                    exit_code = p.wait()
                self.run_until_success(lambda: self.update_exit_code(exit_code))
                while not self.stopped and stdout_writer.is_alive():
                    stdout_writer.join(0.5)
                while not self.stopped and stderr_writer.is_alive():
                    stderr_writer.join(0.5)
        finally:
            if stdin_reader is not None:
                stdin_reader.stop()
                stdin_reader.join()
            if stdin_writer is not None:
                stdin_writer.stop()
                stdin_writer.join()
            if stdout_reader is not None:
                stdout_reader.stop()
                stdout_reader.join()
            if stdout_writer is not None:
                stdout_writer.stop()
                stdout_writer.join()
            if stderr_reader is not None:
                stderr_reader.stop()
                stderr_reader.join()
            if stderr_writer is not None:
                stderr_writer.stop()
                stderr_writer.join()
            if keep_alive is not None:
                keep_alive.stop()
                keep_alive.join()

    def keep_alive(self) -> None:
        requests.post(f"{self.broker_url}/keepAlive?end=host", timeout=5)

    def update_exit_code(self, exit_code: int) -> None:
        resp = requests.post(f"{self.broker_url}?exit_code={exit_code}", timeout=5)
        check_status_code(resp)
        return resp.json()  # type: ignore


class CommandRetriever(TaskRunner):
    def __init__(self, name: str, broker_url: str):
        super().__init__(name)
        self.broker_url = broker_url
        self.executions: Dict[str, CommandExecutor] = {}

    def run(self) -> None:
        try:
            while not self.stopped:
                commands = self.run_until_success(self.retrieve_commands)
                for command in commands:
                    command_id = command["command_id"]
                    args = command["args"]
                    if command_id not in self.executions:
                        print(f"run command {command_id}:", args)
                        executor = CommandExecutor(
                            f"command-{command_id}-executor",
                            f"{self.broker_url}/command/{command_id}",
                            args,
                        )
                        executor.start()
                        self.executions[command_id] = executor
                time.sleep(1)
        finally:
            for executor in self.executions.values():
                executor.stop()
                executor.join()

    def retrieve_commands(self) -> List[Dict[str, Any]]:
        resp = requests.get(f"{self.broker_url}/command", timeout=5)
        check_status_code(resp)
        return resp.json()["commands"]  # type: ignore



from .common import convert_list_to_tensor, convert_numpy_to_tensor
from .dataset import default_transform, TorchIterableDataset

__all__ = [
    "TorchIterableDataset",
    "default_transform",
    "convert_list_to_tensor",
    "convert_numpy_to_tensor",
]



import os
import typing as t
import tarfile
import platform
from abc import ABCMeta, abstractmethod, abstractclassmethod
from pathlib import Path
from contextlib import ExitStack

import yaml
from loguru import logger
from fs.walk import Walker
from fs.tarfs import TarFS

from starwhale.utils import console, now_str, gen_uniq_version
from starwhale.consts import (
    YAML_TYPES,
    CREATED_AT_KEY,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SHORT_VERSION_CNT,
    SW_IGNORE_FILE_NAME,
    DEFAULT_MANIFEST_NAME,
)
from starwhale.version import STARWHALE_VERSION
from starwhale.base.tag import StandaloneTag
from starwhale.utils.fs import move_dir, empty_dir, ensure_dir, ensure_file, extract_tar
from starwhale.utils.venv import SUPPORTED_PIP_REQ
from starwhale.utils.error import FileTypeError, NotFoundError, MissingFieldError
from starwhale.utils.config import SWCliConfigMixed

from .uri import URI
from .store import BundleField


class BaseBundle(metaclass=ABCMeta):
    def __init__(self, uri: URI) -> None:
        self.uri = uri
        self.name = self.uri.object.name
        self.sw_config = SWCliConfigMixed()
        self.yaml_name = ""

    @abstractmethod
    def info(self) -> t.Dict[str, t.Any]:
        raise NotImplementedError

    @abstractmethod
    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        raise NotImplementedError

    @abstractmethod
    def list_tags(self) -> t.List[str]:
        raise NotImplementedError

    @abstractmethod
    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        raise NotImplementedError

    @abstractmethod
    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.List[t.Dict[str, t.Any]]:
        raise NotImplementedError

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filters = filters or {}
        _cls = cls._get_cls(project_uri)
        _filter = cls.get_filter_dict(filters, cls.get_filter_fields())
        return _cls.list(project_uri, page, size, _filter)  # type: ignore

    @classmethod
    def get_filter_dict(
        cls,
        filters: t.Union[t.Dict[str, t.Any], t.List[str]],
        fields: t.Optional[t.List[str]] = None,
    ) -> t.Dict[str, t.Any]:
        fields = fields or []
        if isinstance(filters, t.Dict):
            return {k: v for k, v in filters.items() if k in fields}

        _filter_dict: t.Dict[str, t.Any] = {}
        for _f in filters:
            _item = _f.split("=", 1)
            if _item[0] in fields:
                _filter_dict[_item[0]] = _item[1] if len(_item) > 1 else ""
        return _filter_dict

    @classmethod
    def get_filter_fields(cls) -> t.List[str]:
        return ["name", "owner", "latest"]

    @classmethod
    def do_bundle_filter(
        cls,
        bundle_field: BundleField,
        filters: t.Union[t.Dict[str, t.Any], t.List[str]],
    ) -> bool:
        filter_dict = cls.get_filter_dict(filters, cls.get_filter_fields())
        _name = filter_dict.get("name")
        if _name and not bundle_field.name.startswith(_name):
            return False
        _latest = filter_dict.get("latest") is not None
        if _latest and "latest" not in bundle_field.tags:
            return False

        return True

    @abstractclassmethod
    def _get_cls(cls, uri: URI) -> t.Any:
        raise NotImplementedError

    @classmethod
    def copy(cls, src_uri: str, dest_uri: str) -> None:
        raise NotImplementedError

    def extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        raise NotImplementedError

    def build(self, *args: t.Any, **kwargs: t.Any) -> None:
        self.store.building = True  # type: ignore

        # use a temp dir to build resources
        # and mv results to dst dir to prevent leaving garbage when interrupt
        def when_exit() -> None:
            src = self.store.snapshot_workdir  # type: ignore
            self.store.building = False  # type: ignore
            dst = self.store.snapshot_workdir  # type: ignore
            ensure_dir(dst.parent)
            os.rename(src, dst)
            console.print(f"finish gen resource @ {dst}")

        with ExitStack() as stack:
            stack.callback(when_exit)
            kwargs["yaml_name"] = kwargs.get("yaml_name", self.yaml_name)
            self.buildImpl(*args, **kwargs)

    def buildImpl(self, *args: t.Any, **kwargs: t.Any) -> None:
        raise NotImplementedError

    @property
    def version(self) -> str:
        return getattr(self, "_version", "") or self.uri.object.version


class LocalStorageBundleMixin:
    def __init__(self) -> None:
        self._manifest: t.Dict[str, t.Any] = {}

    def _render_manifest(self) -> None:
        self._manifest["build"] = dict(
            os=platform.system(),
            sw_version=STARWHALE_VERSION,
        )
        self._manifest["version"] = self._version  # type: ignore
        self._manifest[CREATED_AT_KEY] = now_str()

        # TODO: add signature for import files: model, config
        _fpath = self.store.snapshot_workdir / DEFAULT_MANIFEST_NAME  # type: ignore
        ensure_file(_fpath, yaml.safe_dump(self._manifest, default_flow_style=False))
        logger.info(f"[step:manifest]render manifest: {_fpath}")

    def _gen_version(self) -> None:
        logger.info("[step:version]create version...")
        if not getattr(self, "_version", ""):
            self._version = gen_uniq_version()

        self.uri.object.version = self._version  # type:ignore
        logger.info(f"[step:version]version: {self._version}")
        console.print(f":new: version {self._version[:SHORT_VERSION_CNT]}")  # type: ignore

    def _make_auto_tags(self) -> None:
        self.tag.add_fast_tag()  # type: ignore

    def _make_tar(self, ftype: str = "") -> None:
        out = self.store.bundle_dir / f"{self._version}{ftype}"  # type: ignore
        ensure_dir(self.store.bundle_dir)  # type: ignore
        logger.info(f"[step:tar]try to tar {out} ...")

        with tarfile.open(out, "w:") as tar:
            tar.add(str(self.store.snapshot_workdir), arcname="")  # type: ignore

        console.print(f":butterfly: {ftype} bundle:{out}")
        logger.info("[step:tar]finish to make bundle tar")

    @classmethod
    def _do_validate_yaml(cls, path: Path) -> None:
        # add more yaml validators
        if not path.exists():
            raise NotFoundError(str(path))

        if not path.name.endswith(YAML_TYPES):
            raise FileTypeError(f"{path} file type is not yaml|yml")

    def _get_bundle_info(self) -> t.Dict[str, t.Any]:
        _uri = self.uri  # type: ignore
        _store = self.store  # type: ignore

        if not _store.bundle_path.exists():
            return {}

        _manifest: t.Dict[str, t.Any] = {
            "uri": _uri.full_uri,
            "project": _uri.project,
            "name": _uri.object.name,
            "snapshot_workdir": str(_store.snapshot_workdir),
            "bundle_path": str(_store.bundle_path),
        }

        if _uri.object.version:
            _tag = StandaloneTag(_uri)
            _manifest["version"] = _uri.object.version
            _manifest["config"] = {}
            _manifest["tags"] = _tag.list()

            if _store.bundle_path.is_dir():
                _manifest["config"].update(_store.manifest)
            else:
                if _store.snapshot_workdir.exists():
                    _manifest["config"].update(_store.manifest)
                elif _store.bundle_path.exists():
                    with TarFS(str(_store.bundle_path)) as tar:
                        with tar.open(DEFAULT_MANIFEST_NAME) as f:
                            _om = yaml.safe_load(f)
                        _manifest["config"].update(_om)
                else:
                    raise NotFoundError(
                        f"{_store.bundle_path} and {_store.snapshot_workdir}"
                    )
        else:
            _manifest["history"] = self.history()  # type: ignore
        return _manifest

    def _do_extract(self, force: bool = False, target: t.Union[str, Path] = "") -> Path:
        _store = self.store  # type: ignore
        _uri = self.uri  # type: ignore

        if not _uri.object.version:
            raise MissingFieldError("no version")

        _target: Path = Path(target) if target else _store.snapshot_workdir

        if (
            _target.exists()
            and (_target / DEFAULT_MANIFEST_NAME).exists()
            and not force
        ):
            console.print(f":joy_cat: {_target} existed, skip extract model bundle")
        else:
            extract_tar(_store.bundle_path, _target, force)

        return _target

    def _get_src_walker(
        self,
        workdir: Path,
        include_files: t.Optional[t.List[str]] = None,
        exclude_files: t.Optional[t.List[str]] = None,
    ) -> Walker:
        include_files = include_files or []
        exclude_files = exclude_files or []
        _filter = (
            ["*.py", "*.sh", "*.yaml", "*.yml"] + SUPPORTED_PIP_REQ + include_files
        )
        _exclude = exclude_files

        _sw_ignore_path = workdir / SW_IGNORE_FILE_NAME
        if _sw_ignore_path.exists():
            for _l in _sw_ignore_path.read_text().splitlines():
                _l = _l.strip()
                if not _l or _l.startswith("#"):
                    continue

                _exclude.append(_l)

        # Notice: if pass [] as exclude_dirs value, walker will failed
        _exclude = _exclude or None  # type: ignore
        return Walker(filter=_filter, exclude_dirs=_exclude)

    def _do_remove(self, force: bool = False) -> t.Tuple[bool, str]:
        from .store import BaseStorage

        store: BaseStorage = self.store  # type: ignore

        if force:
            empty_dir(store.loc)
            empty_dir(store.snapshot_workdir)
            return True, ""
        else:
            _ok, _reason = move_dir(store.loc, store.recover_loc, False)
            _ok2, _reason2 = True, ""
            if store.snapshot_workdir.exists():
                _ok2, _reason2 = move_dir(
                    store.snapshot_workdir,
                    store.recover_snapshot_workdir,
                    False,
                )
            return _ok and _ok2, _reason + _reason2



from ._impl.evaluation import predict, evaluate, PipelineHandler, EvaluationLogStore

log = EvaluationLogStore.log
log_summary = EvaluationLogStore.log_summary
iter = EvaluationLogStore.iter

__all__ = [
    "PipelineHandler",
    "predict",
    "evaluate",
    "log",
    "log_summary",
    "iter",
]



import sys
from pathlib import Path

import click

from starwhale.utils import console, get_current_shell
from starwhale.utils.fs import ensure_dir
from starwhale.utils.cli import AliasedGroup

_SUPPORT_SHELLS = {
    "bash": ("~/.bashrc", 'eval "$(_{prog_u}_COMPLETE=bash_source {prog})"'),
    "zsh": ("~/.zshrc", 'eval "$(_{prog_u}_COMPLETE=zsh_source {prog})"'),
    "fish": (
        "~/.config/fish/completions/{prog}.fish",
        "eval (env _{prog_u}_COMPLETE=fish_source {prog})",
    ),
}


@click.group(
    "completion", cls=AliasedGroup, help="Shell completion for Starwhale command-line"
)
def completion_cmd() -> None:
    pass


@completion_cmd.command("show", help="Show shell completion code")
@click.argument("shell", required=False, type=click.Choice(_SUPPORT_SHELLS.keys()))
def show(shell: str) -> None:
    _inject_code(shell, stdout=True)


@completion_cmd.command(
    "install", help="Install shell completion script in the current shell"
)
@click.argument("shell", required=False, type=click.Choice(_SUPPORT_SHELLS.keys()))
def completion(shell: str) -> None:
    _inject_code(shell)


def _inject_code(shell: str = "", stdout: bool = False) -> None:
    prog: str = click.get_current_context().find_root().info_name  # type: ignore
    shell = shell or get_current_shell()

    if shell not in _SUPPORT_SHELLS:
        console.print(f":crying_face: no support {shell} completion")
        sys.exit(1)

    _path, code = _SUPPORT_SHELLS[shell]
    path = Path(_path.format(prog=prog)).expanduser().absolute()
    code = code.format(prog_u=prog.upper(), prog=prog)

    if stdout:
        console.print(code)
    else:
        ensure_dir(path.parent)

        mode = "a" if path.exists() else "w"
        with path.open(mode=mode) as f:
            f.write(f"\n#Generated by Starwhale for {prog} completion\n")
            f.write(code + "\n")

        console.print(f":clap: {prog} {shell} completion installed in {path}")
        console.print(
            f":beer: run [bold magenta]exec {shell}[/] command to activate shell completion"
        )



from __future__ import annotations

import typing as t
from abc import abstractmethod
from dataclasses import dataclass

from rich import console
from rich.text import Text
from rich.style import Style
from rich.table import box, Table
from textual.app import Reactive
from textual.widget import Widget

from starwhale.utils import (
    Order,
    get_field,
    pretty_bytes,
    sort_obj_list,
    snake_to_camel,
)
from starwhale.consts import CREATED_AT_KEY, DEFAULT_PROJECT, STANDALONE_INSTANCE
from starwhale.core.eval.view import JobTermView
from starwhale.core.model.view import ModelTermView
from starwhale.core.dataset.view import DatasetTermView
from starwhale.core.runtime.view import RuntimeTermView

if t.TYPE_CHECKING:
    from textual import events
    from textual.widget import RenderableType


default_project = f"{STANDALONE_INSTANCE}/{DEFAULT_PROJECT}"


@dataclass
class Column:
    key: str
    name: t.Optional[str] = ""
    render: t.Optional[t.Callable[[int, t.Any], t.Any]] = None


class OrderBy:
    def __init__(self) -> None:
        self.orderby_keys: t.Dict[str, str] = {
            "C": CREATED_AT_KEY,
            "N": "name",
            "S": "size",
        }
        self.current_order = Order("")

    def record_key(self, key: str) -> bool:
        if key not in self.orderby_keys:
            return False

        field = self.orderby_keys[key]
        if self.current_order.field == field:
            self.current_order.reverse = not self.current_order.reverse
        else:
            self.current_order.field = field
            self.current_order.reverse = False
        return True

    def sort(self, data: t.Sequence) -> t.Sequence:
        if not self.current_order.field:
            return data
        return sort_obj_list(data, [self.current_order])

    def get_order_icon(self) -> t.Tuple[str, RenderableType]:
        if not self.current_order.field:
            return "", ""
        field = self.current_order.field
        if self.current_order.reverse:
            return field, Text(" ↓", style="green")
        else:
            return field, Text(" ↑", style="red")


class TableWidget(Widget):
    """TableWidget makes an interactive rich.Table"""

    def __init__(self, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.table = Table(expand=True, box=box.SIMPLE)
        self.data: t.Sequence = []
        self.render_fn: t.List[Column] = []
        self._info: t.Any = None
        self._orderby = OrderBy()

    show_info: Reactive[bool] = Reactive(False)
    cursor_line: Reactive[int] = Reactive(0, repaint=False)

    def watch_show_info(self, show: bool) -> None:
        self._info = show and self.info(self.cursor_line) or None
        self.refresh(layout=True)

    def watch_cursor_line(self, value: int) -> None:
        self.highlight_row(value)
        self.refresh()

    @abstractmethod
    def reloadImpl(self) -> None:
        raise NotImplementedError

    def info(self, idx: int) -> RenderableType:
        return console.Pretty(self.data[idx], indent_guides=True)

    def watch_data(self) -> None:
        pass

    def render(self) -> Table:
        self.app.sub_title = self.__class__.__name__
        return self._info or self.table

    def reload(self) -> None:
        self.table.columns = []
        for i in self.render_fn:
            name = Text(i.name and i.name or snake_to_camel(i.key))
            f, icon = self._orderby.get_order_icon()
            if i.key == f:
                name += icon
            self.table.add_column(name)
        self.reloadImpl()
        self.table.rows = []

        data = self._orderby.sort(self.data)
        for idx, item in enumerate(data):

            def try_render(col: Column) -> t.Any:
                if col.render:
                    return col.render(idx, item)
                return get_field(item, col.key)

            self.table.add_row(*[try_render(i) for i in self.render_fn])

        self.highlight_row(self.cursor_line)
        self.refresh()

    def highlight_row(self, row: int) -> None:
        self.table.row_styles = [
            Style(bgcolor="magenta") if i == row else ""
            for i in range(self.table.row_count)
        ]

    async def on_key(self, event: events.Key) -> None:
        if event.key == "r":
            self.reload()
        if self._orderby.record_key(event.key):
            self.reload()
        await self.dispatch_key(event)

    async def key_down(self) -> None:
        self.cursor_down()

    async def key_j(self) -> None:
        self.cursor_down()

    async def key_up(self) -> None:
        self.cursor_up()

    async def key_k(self) -> None:
        self.cursor_up()

    def cursor_down(self) -> None:
        if self.cursor_line < self.table.row_count - 1:
            self.cursor_line += 1

    def cursor_up(self) -> None:
        if self.cursor_line > 0:
            self.cursor_line -= 1

    async def key_i(self) -> None:
        self.show_info = True

    async def key_escape(self) -> None:
        self.show_info = False

    async def key_h(self) -> None:
        self.show_info = False


class Models(TableWidget):
    """Models represents starwhale model view"""

    # TODO use constance
    def __init__(self, uri: str = default_project, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.render_fn = [
            Column("name"),
            Column("version"),
            Column("tags", render=lambda _, x: ",".join(x["tags"])),
            Column("size", render=lambda _, x: pretty_bytes(x["size"])),
            Column(CREATED_AT_KEY, "Created At"),
        ]
        self.uri = uri
        self.reload()

    def reloadImpl(self) -> None:
        self.data, _ = ModelTermView.list(self.uri)


class Datasets(TableWidget):
    """Datasets represents starwhale model view"""

    def __init__(self, uri: str = default_project, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.render_fn = [
            Column("name"),
            Column("version"),
            Column("tags", render=lambda _, x: ",".join(x["tags"])),
            Column("size", render=lambda _, x: pretty_bytes(x["size"])),
            Column(CREATED_AT_KEY, "Created At"),
        ]
        self.uri = uri
        self.reload()

    def reloadImpl(self) -> None:
        self.data, _ = DatasetTermView.list(self.uri)


class Runtimes(TableWidget):
    """Runtimes represents starwhale model view"""

    def __init__(self, uri: str = default_project, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.render_fn = [
            Column("name"),
            Column("version"),
            Column("tags", render=lambda _, x: ",".join(x["tags"])),
            Column("size", render=lambda _, x: pretty_bytes(x["size"])),
            Column(CREATED_AT_KEY, "Created At"),
        ]
        self.uri = uri
        self.reload()

    def reloadImpl(self) -> None:
        self.data, _ = RuntimeTermView.list(self.uri)


class Jobs(TableWidget):
    """Job represents starwhale model view"""

    def __init__(self, uri: str = default_project, **kwargs: t.Any) -> None:
        super().__init__(**kwargs)
        self.render_fn = [
            Column("manifest.version", "Name"),
            Column("manifest.model", "Model"),
            Column(
                "manifest.datasets",
                "Datasets",
                render=lambda _, x: console.Pretty(get_field(x, "manifest.datasets")),
            ),
            Column(f"manifest.{CREATED_AT_KEY}", "Created At"),
            Column("manifest.finished_at", "Finished At"),
        ]
        self.uri = uri
        self.reload()

    def reloadImpl(self) -> None:
        self.data, _ = JobTermView.list(self.uri)



import sys
import typing as t
from http import HTTPStatus

import requests
from rich import box
from rich.panel import Panel
from rich.table import Table

from starwhale.utils import console, fmt_http_server
from starwhale.consts import UserRoleType, SW_API_VERSION, STANDALONE_INSTANCE
from starwhale.base.view import BaseTermView
from starwhale.utils.http import wrap_sw_error_resp

from .model import CloudInstance

DEFAULT_HTTP_TIMEOUT = 90


class InstanceTermView(BaseTermView):
    def __init__(self) -> None:
        super().__init__()

    def select(self, instance: str) -> None:
        try:
            self.select_current_default(instance=instance)
        except Exception as e:
            console.print(
                f":person_shrugging: failed to select {instance}, reason: {e}"
            )
            sys.exit(1)
        else:
            console.print(f":clap: select {self.current_instance} instance")

    def login(self, instance: str, alias: str, **kw: str) -> None:
        if instance == STANDALONE_INSTANCE:
            console.print(f":pinching_hand: skip {instance} instance login")
            return

        instance = instance or self.sw_remote_addr
        server = fmt_http_server(instance)
        if kw.get("token"):
            r = self._login_request_by_token(server, kw["token"])
        else:
            r = self._login_request_by_username(server, kw)

        if r.status_code == HTTPStatus.OK:
            console.print(f":man_cook: login {server} successfully!")
            token = r.headers.get("Authorization") or kw.get("token")
            if not token:
                console.print("cannot get token, please contract starwhale")
                sys.exit(1)

            _d = r.json()["data"]
            _role = _d.get("role", {}).get("roleName") if isinstance(_d, dict) else None

            self.update_instance(
                uri=server,
                user_name=_d.get("name", ""),
                user_role=_role or UserRoleType.NORMAL,
                sw_token=token,
                alias=alias,
            )
        else:
            wrap_sw_error_resp(r, "login failed!", exit=True)

    def _login_request_by_token(self, server: str, token: str) -> requests.Response:
        url = f"{server}/api/{SW_API_VERSION}/user/current"
        return requests.get(
            url,
            timeout=DEFAULT_HTTP_TIMEOUT,
            verify=False,
            headers={"Authorization": token},
        )

    def _login_request_by_username(
        self, server: str, auth_request: t.Dict[str, str]
    ) -> requests.Response:
        url = f"{server}/api/{SW_API_VERSION}/login"
        return requests.post(
            url,
            verify=False,
            timeout=DEFAULT_HTTP_TIMEOUT,
            data={
                "userName": auth_request["username"],
                "userPwd": auth_request["password"],
            },
        )

    def logout(self, instance: str = "") -> None:
        # TODO: do real logout request
        instance = instance or self.current_instance

        if instance == STANDALONE_INSTANCE:
            console.print(f":pinching_hand: skip {instance} instance logout")
            return
        self.delete_instance(instance)
        console.print(":wink: bye.")

    def info(self, instance: str = "") -> t.Dict:
        instance = instance or self.current_instance

        if instance == STANDALONE_INSTANCE:
            return {
                "instance": instance,
                "root_dir": str(self.rootdir),
            }
        else:
            # TODO: support use uri directly
            # TODO: user async to get
            ci = CloudInstance(instance)
            _version = ci._fetch_version()
            _agents = ci._fetch_agents()

            result = {
                "instance": instance,
                "version": _version,
                "agents": [
                    {"ip": i["ip"], "status": str(i["status"]), "version": i["version"]}
                    for i in _agents
                ],
            }
            return result

    def list(self) -> t.List[t.Dict[str, t.Any]]:
        result = list()

        for k, v in self._config["instances"].items():
            _is_current = (
                k == self.current_instance or v["uri"] == self.current_instance
            )
            result.append(
                {
                    "in_use": _is_current,
                    "name": k,
                    "uri": v.get("uri", ""),
                    "user_name": v.get("user_name", ""),
                    "user_role": v.get("user_role", ""),
                    "current_project": str(v.get("current_project", "")),
                    "updated_at": v.get("updated_at", ""),
                }
            )
        return result


class InstanceTermViewRich(InstanceTermView):
    def list(self) -> None:  # type: ignore
        title = "List Starwhale Instances"
        custom_table = {"caption": f"Current Instance: [blink]{self.current_instance}"}
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "in_use": lambda x: ":backhand_index_pointing_right:" if x else "",
            "user_role": self.place_holder_for_empty(),
            "current_project": self.place_holder_for_empty(),
            "updated_at": self.place_holder_for_empty(),
        }
        custom_row = lambda row: {"style": "magenta"} if row["in_use"] else None
        data = super().list()
        self.print_table(
            title,
            data,
            custom_table=custom_table,
            custom_column=custom_column,
            custom_row=custom_row,
        )

    @BaseTermView._header  # type: ignore
    def info(self, instance: str = "") -> None:
        _info = super().info(instance)
        instance = _info["instance"]

        if instance == STANDALONE_INSTANCE:
            console.print(f":balloon: standalone instance, root dir @ {self.rootdir}")
        else:

            def _agents_table() -> Table:
                table = Table(
                    show_edge=False,
                    show_header=True,
                    box=box.SIMPLE,
                )
                table.add_column("id")
                table.add_column("ip", style="green")
                table.add_column("status", style="blue")
                table.add_column("version")

                for i, _agent in enumerate(_info["agents"]):
                    table.add_row(
                        str(i),
                        _agent["ip"],
                        _agent["status"],
                        _agent["version"],
                    )
                return table

            def _details() -> Panel:
                grid = Table.grid(padding=1, pad_edge=True)
                grid.add_column(
                    "Category", no_wrap=True, justify="left", style="bold green"
                )
                grid.add_column("Information")
                grid.add_row("Version", _info["version"])
                grid.add_row(
                    "Agents",
                    _agents_table(),
                )

                return Panel(grid, title_align="left")

            console.print(_details())


class InstanceTermViewJson(InstanceTermView):
    def list(self) -> None:  # type: ignore
        self.pretty_json(super().list())

    def info(self, instance: str = "") -> None:  # type: ignore
        self.pretty_json(super().info(instance))


def get_term_view(ctx_obj: t.Dict) -> t.Type[InstanceTermView]:
    return (
        InstanceTermViewJson
        if ctx_obj.get("output") == "json"
        else InstanceTermViewRich
    )



from __future__ import annotations

import os
import copy
import json
import time
import queue
import atexit
import typing as t
import threading
from types import TracebackType
from pathlib import Path

import yaml
from loguru import logger

from starwhale.utils import now_str, random_str, gen_uniq_version
from starwhale.consts import SW_AUTO_DIRNAME, DEFAULT_MANIFEST_NAME
from starwhale.base.uri import URI
from starwhale.utils.fs import ensure_dir, ensure_file
from starwhale.base.type import URIType, InstanceType
from starwhale.consts.env import SWEnv
from starwhale.utils.dict import flatten
from starwhale.utils.error import NoSupportError
from starwhale.utils.config import SWCliConfigMixed
from starwhale.core.dataset.type import BaseArtifact

from .base import (
    _TrackMode,
    _TrackSource,
    ParamsRecord,
    MetricsRecord,
    ArtifactsRecord,
    HandleQueueElementType,
)
from .hooker import ConsoleLogHook, hook_frameworks, hook_python_libs
from .syncer import SyncerThread
from .handler import HandlerThread

_INSTANCE_NOT_INIT_ERROR = (
    "Tracker instance has not been initialized, please call track.start() at first"
)


# TODO: use multi-process model to refactor Tracer for GIL performance
class Tracker:
    _instance: t.Optional[Tracker] = None
    _lock = threading.Lock()

    def __init__(
        self,
        name: str,
        project_uri: URI,
        access_token: str = "",
        auto_run_collect: bool = True,
        auto_framework_hook: bool = True,
        auto_console_log_redirect: bool = True,
        mode: _TrackMode = _TrackMode.ONLINE,
        with_id: str = "",
        tags: t.Optional[t.List[str]] = None,
        description: str = "",
        collector_sample_interval: float = 1.0,
        collector_report_interval: float = 30.0,
        saved_dir: t.Optional[t.Union[str, Path]] = None,
    ) -> None:
        name = name.strip()
        if not name:
            raise ValueError("name field is empty")
        self.name = name

        if not project_uri.project:
            raise ValueError(
                f"{project_uri} is wrong format, that cannot fetch project field"
            )

        if project_uri.instance_type == InstanceType.CLOUD:
            access_token = (
                access_token
                or SWCliConfigMixed().get_sw_token(instance=project_uri.instance)
                or os.environ.get(SWEnv.instance_token, "")
            )
            if not access_token:
                raise ValueError("access_token is required for cloud instance")

        self.project_uri = project_uri
        self.access_token = access_token
        self.auto_run_collect = auto_run_collect
        self.auto_framework_hook = auto_framework_hook
        self.auto_console_log_redirect = auto_console_log_redirect
        # TODO: support envs setting for collector_*_interval
        self.collector_sample_interval = collector_sample_interval
        self.collector_report_interval = collector_report_interval
        self.mode = mode
        self.tags = tags or []
        self.description = description

        if saved_dir is None:
            saved_dir = Path.cwd()

        self._saved_dir = Path(saved_dir).absolute()

        self._id = with_id or gen_uniq_version()
        # TODO: support with_id argument from existed Tracker in the local storage
        self._metric_step = 0
        self._metric_lock = threading.Lock()
        self._uncommitted_metrics: t.Dict[str, float] = {}

        self._handle_queue: queue.Queue[HandleQueueElementType] = queue.Queue()
        self._sync_queue: queue.Queue[t.Any] = queue.Queue()
        self._threads: t.List[threading.Thread] = []
        self._handler_thread: t.Optional[threading.Thread] = None
        self._syncer_thread: t.Optional[threading.Thread] = None
        self._collector_thread: t.Optional[threading.Thread] = None

        self._start_time = time.monotonic()
        self._workdir: t.Optional[Path] = None
        self._log_hook: t.Optional[ConsoleLogHook] = None
        self._manifest: t.Dict[str, t.Any] = {"start_time": now_str()}

    def __str__(self) -> str:
        return f"Starwhale Tracker[{self.mode.value}]-{self.name}-{self._id}"

    def __repr__(self) -> str:
        return f"Starwhale Tracker[{self.mode.value}]-{self.name}-{self._id}-{self.project_uri}"

    def __enter__(self) -> Tracker:
        return self

    def __exit__(
        self,
        type: t.Optional[t.Type[BaseException]],
        value: t.Optional[BaseException],
        trace: TracebackType,
    ) -> None:
        if value:  # pragma: no cover
            logger.warning(f"type:{type}, exception:{value}, traceback:{trace}")

        self.end()

    @property
    def id(self) -> str:
        return self._id

    # TODO: support re-init for Tracker, create multi trackers
    @classmethod
    def start(
        cls,
        name: str = "",
        project_uri: t.Optional[t.Union[str, URI]] = None,
        access_token: str = "",
        auto_run_collect: bool = True,
        auto_framework_hook: bool = True,
        auto_console_log_redirect: bool = True,
        mode: str = "online",
        with_id: str = "",
        tags: t.Optional[t.List[str]] = None,
        description: str = "",
        collector_sample_interval: float = 1.0,
        collector_report_interval: float = 30.0,
        saved_dir: t.Optional[t.Union[str, Path]] = None,
    ) -> Tracker:
        """Starts an new tracker to track metrics, artifacts and parameters to Starwhale.

        In your python script code, you should add `track.start()` to the beginning of training or evaluation code block,
        and every thing will be tracked to Starwhale.

        Arguments:
            name: (str, optional) The name of tracker. If the name is not specified,
                a random and human-readable name will be generated. Every name will be added an increment index suffix automatically.
            project_uri: (str, URI, optional) The project URI is the place where tracker sends records.
                If the project_uri is not specified, the default selected instance and project will be used. Standalone or Cloud instance
                project is ok.
            access_token: (str, optional) When project_uri is the cloud project, access_token is required. If access_token is not specified,
                starwhale will try to fetch token by `SW_TOKEN` environment or `~/.config/starwhale/config.yaml` config file.
            auto_run_collect: (bool, optional) When auto_environment value is `True`, Tracker will log program run environment information
                automatically. The run environment information includes cpu/gpu/mem/python/os and so on. `auto_environment` is `True` in default.
            auto_framework_hook: (bool, optional) Auto record MachineLearning Frameworks metrics. `auto_framework` is `True` in default.
            auto_console_log_redirect: (bool, optional) Tracker will catch program stdout and stderr console log. `auto_console_log` is `True` in default.
            mode: (str, optional) Config track mode. Options: `online` or `offline`. Defaults to `online`.
                cases:
                - `online`: sync track data to Starwhale Server.
                - `offline`: Only record track data in the local dir. You can run `swcli track track` manually.
            with_id: (str, optional) Reuse the existed Tracker. id is the tracker uuid.
            tags: (list[str], optional) Tags are used to organize, filter tracks.
            description: (str, optional) Tacker description.
            collector_sample_interval: (float, optional) the interval seconds of the run collector collect, default is 2.0 seconds.
            collector_report_interval: (float, optional) the interval seconds of the run collector report, all values will be aggregated to one value. default is 30.0 seconds.
            saved_dir: (str, pathlib.Path, optional) the saved dir for track, default is cwd dir.

        Examples:
        ```python
        from starwhale import track
        track.start()
        track.metrics(loss=0.768, accuracy=0.9)
        track.end()

        Returns:
            A `Tracker` object.
        ```
        """
        with Tracker._lock:
            if Tracker._instance is not None:
                raise RuntimeError(f"{Tracker._instance} has already been initialized")

            if project_uri is None:
                scm = SWCliConfigMixed()
                _inst = scm.current_instance
                _prj = scm.current_project
                if not _inst or not _prj:
                    raise ValueError(
                        f"cannot get default project uri(instance:{_inst}, project:{_prj}) from swcli config"
                    )
                project_uri = URI(
                    f"{_inst}/project/{_prj}", expected_type=URIType.PROJECT
                )
            elif isinstance(project_uri, str):
                project_uri = URI(project_uri, expected_type=URIType.PROJECT)
            elif isinstance(project_uri, URI):
                project_uri = project_uri
            else:
                raise TypeError(
                    f"project_uri({project_uri}) type is unexpected, only accepts str, URI or None"
                )

            name = name.strip()
            if not name:
                name = random_str()

            _tracker = Tracker(
                name=name,
                project_uri=project_uri,
                access_token=access_token,
                auto_run_collect=auto_run_collect,
                auto_framework_hook=auto_framework_hook,
                auto_console_log_redirect=auto_console_log_redirect,
                mode=_TrackMode(mode),
                with_id=with_id,
                tags=tags,
                description=description,
                collector_sample_interval=collector_sample_interval,
                collector_report_interval=collector_report_interval,
                saved_dir=saved_dir,
            )
            _tracker._setup()
            Tracker._instance = _tracker
            return _tracker

    @classmethod
    def end(cls) -> None:
        """Ends a track.
        Singleton Tracker instance will be released. If users don't call end function,
        Tracker will call end function when the process exiting automatically.
        """
        with Tracker._lock:
            if Tracker._instance is None:
                return

            _inst = Tracker._instance
            Tracker._instance = None

            _inst._render_manifest()
            _inst._close_threads()

            if _inst._log_hook is not None:
                _inst._log_hook.uninstall()
                _inst._log_hook = None

            atexit.unregister(Tracker.end)

    @classmethod
    def metrics(
        cls,
        *args: t.Dict,
        step: t.Optional[int] = None,
        commit: bool = True,
        **kwargs: float,
    ) -> Tracker:
        """Track metrics to Starwhale.
        Metric value type must be float or convertible-float type. Each call to `metrics` function will trigger
        internal step counter to be added by 1.

        Arguments:
            *args: [dict, optional] use dicts to store metrics.
            **kwargs: [float, optional] use kv to store metrics.
            step: [int, optional] custom step value.
            commit: [bool, optional] when `commit` is `false`, the metrics will be aggregated with the same step value
                until the next `track.metrics(commit=True)` function call. Defaults to `true`.

        Examples:
        ```python
        track.metrics(loss=0.99)
        track.metrics(loss=0.99, accuracy=0.98)
        track.metrics({"loss": 0.99, "accuracy": 0.98})

        # loss and accuracy use the same step.
        track.metrics(loss=0.99, commit=False)
        track.metrics(accuracy=0.98)

        # use custom step value
        track.metrics(loss=0.99, step=1)

        Returns:
            A `Tracker` object.
        ```
        """
        if cls._instance is None:
            raise RuntimeError(_INSTANCE_NOT_INIT_ERROR)

        data = cls._merge_dicts(args, kwargs)
        cls._instance._log_metrics(data=data, step=step, commit=commit)
        return cls._instance

    def _log_metrics(
        self,
        data: t.Dict[str, float],
        step: t.Optional[int] = None,
        commit: bool = False,
        source: _TrackSource = _TrackSource.USER,
    ) -> None:
        if not data:
            return

        _data = flatten(data, extract_sequence=True)
        for k, v in _data.items():
            _data[k] = float(v)

        _step: int = self._metric_step
        if step is not None:
            if not isinstance(step, int) or step < 0:
                raise TypeError(
                    f"step({step}) is unexpected, the normal step is non-negative integer"
                )
            _step = step

        with self._metric_lock:
            self._uncommitted_metrics.update(_data)
            if commit:
                self._handle_queue.put(
                    MetricsRecord(
                        data=copy.deepcopy(self._uncommitted_metrics),
                        step=_step,
                        source=source,
                        start_time=self._start_time,
                    )
                )
                self._metric_step = _step + 1
                self._uncommitted_metrics = {}

    @classmethod
    def _merge_dicts(cls, *args: t.Any, **kwargs: t.Any) -> t.Dict:
        data: t.Dict = {}
        for item in args:
            if not item:
                continue
            elif isinstance(item, tuple):
                for i in item:
                    if i and isinstance(i, dict):
                        data.update(i)
            elif isinstance(item, dict):
                data.update(item)
            else:
                raise TypeError(f"item{item} no support type")

        data.update(kwargs)
        return data

    @classmethod
    def params(cls, *args: t.Dict[str, t.Any], **kwargs: t.Any) -> Tracker:
        """Track parameters to Starwhale.
        We can use `track.params` to record hyperparameter and configuration. The arguments must be jsonable values.

        Arguments:
            *args: [dict, optional] use dicts to store parameters.
            **kwargs: [optional] use kv to store parameters.

        Examples:
        ```python
        track.params(model="multi-classification")
        track.params({"batch_size": 10})
        ```
        Returns:
            A `Tracker` object.
        """
        if cls._instance is None:
            raise RuntimeError(_INSTANCE_NOT_INIT_ERROR)
        data = cls._merge_dicts(args, kwargs)
        cls._instance._log_params(data)
        return cls._instance

    def _log_params(
        self, data: t.Dict[str, t.Any], source: _TrackSource = _TrackSource.USER
    ) -> None:
        if not data:
            return

        if data != json.loads(json.dumps(data)):
            raise TypeError(f"{data} keys should be string type")

        self._handle_queue.put(
            ParamsRecord(data=data, source=source, start_time=self._start_time)
        )

    @classmethod
    def artifacts(cls, *args: t.Dict[str, t.Any], **kwargs: t.Any) -> Tracker:
        """Track artifacts to Starwhale.

        Arguments:
            *args: [dict, optional] use dicts to store artifacts.
            **kwargs: [optional] use kv to store artifacts.

        Examples:
        ```python
        track.artifacts(model=Model("./model/mnist.pth"))
        track.artifacts(image=Image("sample.png"))
        ```
        Returns:
            A `Tracker` object.
        """
        if cls._instance is None:
            raise RuntimeError(_INSTANCE_NOT_INIT_ERROR)
        data = cls._merge_dicts(args, kwargs)
        cls._instance._log_artifacts(data)
        return cls._instance

    def _log_artifacts(
        self, data: t.Dict[str, t.Any], source: _TrackSource = _TrackSource.USER
    ) -> None:
        if not data:
            return

        _data = flatten(data, extract_sequence=True)
        for k, v in _data.items():
            # TODO: support more artifacts type
            if not isinstance(v, BaseArtifact):
                raise NoSupportError(f"v({v}:{type(v)}) not support for artifacts")

        self._handle_queue.put(
            ArtifactsRecord(data=_data, source=source, start_time=self._start_time)
        )

    def _close_threads(self) -> None:
        for _t in self._threads:
            if _t is None:
                continue
            _t.close()  # type: ignore

    def _setup_dirs(self) -> None:
        track_rootdir = self._saved_dir / SW_AUTO_DIRNAME / "track"
        ensure_dir(track_rootdir)

        target_idx = -1
        for d in track_rootdir.iterdir():
            if not d.is_dir() or "-" not in d.name:
                continue

            idx, tid = d.name.split("-", 1)
            if tid == self._id:
                self._workdir = track_rootdir / d.name
                return

            target_idx = max(target_idx, int(idx))

        self._manifest["track_incr_index"] = target_idx + 1
        self._workdir = track_rootdir / f"{target_idx + 1}-{self._id}"
        ensure_dir(self._workdir)
        ensure_dir(self._workdir / "logs")
        ensure_dir(self._workdir / "metrics")
        ensure_dir(self._workdir / "params")
        ensure_dir(self._workdir / "artifacts")

    def _setup(self) -> None:
        atexit.register(Tracker.end)
        self._setup_dirs()

        if self._workdir is None:
            raise RuntimeError("workdir is None")

        if self.auto_run_collect:
            from .collector import CollectorThread

            self._collector_thread = CollectorThread(
                tracker=self,
                sample_interval=self.collector_sample_interval,
                report_interval=self.collector_report_interval,
                workdir=self._workdir,
            )
            self._threads.append(self._collector_thread)

        if self.mode == _TrackMode.ONLINE:
            self._syncer_thread = SyncerThread(self._sync_queue, self.project_uri)
            self._threads.append(self._syncer_thread)
            _sync_queue = self._sync_queue
        else:
            _sync_queue = None

        self._handler_thread = HandlerThread(
            workdir=self._workdir,
            handle_queue=self._handle_queue,
            sync_queue=_sync_queue,
        )
        self._threads.append(self._handler_thread)

        for _t in self._threads:
            if _t is None:  # pragma: no cover
                continue
            _t.start()

        if self.auto_console_log_redirect and self._log_hook is None:
            self._log_hook = ConsoleLogHook(
                log_dir=self._workdir / "logs",
                track_id=self._id,
            )
            self._log_hook.install()

        if self.auto_framework_hook:
            hook_frameworks()
            hook_python_libs()

    def _render_manifest(self) -> None:
        if self._workdir is None:
            raise RuntimeError("can not fetch workdir")

        self._manifest.update(
            {
                "id": self._id,
                "name": self.name,
                "project_uri": str(self.project_uri),
                "end_time": now_str(),
                "auto": {
                    "run_collect": self.auto_run_collect,
                    "console_log_redirect": self.auto_console_log_redirect,
                    "framework_hook": self.auto_framework_hook,
                },
                "collector_interval": {
                    "sample": self.collector_sample_interval,
                    "report": self.collector_report_interval,
                },
                "mode": self.mode.value,
                "tags": self.tags,
                "description": self.description,
                "workdir": str(self._workdir),
            }
        )

        ensure_dir(self._workdir)
        ensure_file(
            self._workdir / DEFAULT_MANIFEST_NAME,
            yaml.safe_dump(self._manifest, default_flow_style=False),
        )



from fastapi import APIRouter

from starwhale.web.user import user
from starwhale.web.response import success, SuccessResp

router = APIRouter()
prefix = "project"

project = {
    "id": "1",
    "name": "self",
    "privacy": "PRIVATE",
    "createdTime": 0,
    "owner": user,
    "statistics": {},
}


@router.get("/{project_id}")
def get_project(project_id: str) -> SuccessResp:
    return success(project)


@router.get("/{project_id}/job/{job}")
def get_job(project_id: str, job: str) -> SuccessResp:
    return success({"id": "1", "uuid": job, "modelName": "mock-model"})


@router.get("/{project_id}/role")
def get_role(project_id: str) -> SuccessResp:
    return success(
        [
            {
                "id": "1",
                "user": user,
                "project": project,
                "role": {
                    "id": "1",
                    "name": "Owner",
                    "code": "OWNER",
                },
            },
        ]
    )



import sys
import typing as t

from rich.tree import Tree
from rich.panel import Panel
from rich.pretty import Pretty

from starwhale.utils import console, pretty_bytes
from starwhale.consts import (
    CREATED_AT_KEY,
    DEFAULT_PROJECT,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
)
from starwhale.base.uri import URI
from starwhale.base.type import URIType
from starwhale.base.view import BaseTermView

from .model import Project, ProjectObjType


class ProjectTermView(BaseTermView):
    def __init__(self, project_uri: str = "") -> None:
        super().__init__()
        self.raw_uri = project_uri
        self.uri = URI(project_uri, expected_type=URIType.PROJECT)
        self.project = Project.get_project(self.uri)

    @BaseTermView._simple_action_print
    def create(self) -> t.Tuple[bool, str]:
        return self.project.create()

    @classmethod
    def list(
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        projects, pager = Project.list(instance_uri, page, size)

        _current_project = URI(
            instance_uri, expected_type=URIType.INSTANCE
        ).sw_instance_config.get("current_project", DEFAULT_PROJECT)

        result = list()
        for _p in projects:
            _name = _p["name"]
            _owner = _p.get("owner", "")
            _c_project, _c_owner = URI.uri_to_project_and_owner(_current_project)
            _is_current = _name == _c_project and (_c_owner == "" or _owner == _c_owner)

            result.append(
                {
                    "in_use": _is_current,
                    "name": _name,
                    "location": _p.get("location", ""),
                    "owner": _owner,
                    CREATED_AT_KEY: _p[CREATED_AT_KEY],
                }
            )

        return result, pager

    def select(self) -> None:
        try:
            self.select_current_default(
                instance=self.uri.instance_alias or self.uri.instance,
                project=self.uri.project,
            )
        except Exception as e:
            console.print(
                f":broken_heart: failed to select {self.raw_uri}, reason: {e}"
            )
            sys.exit(1)
        else:
            console.print(
                f":clap: select instance:{self.current_instance}, project:{self.current_project} successfully"
            )

    def remove(self, force: bool = False) -> None:
        ok, reason = self.project.remove(force)
        if ok:
            console.print(
                f":dog: remove project {self.project.name}, you can recover it, don't panic."
            )
        else:
            console.print(
                f":fearful: failed to remove project {self.project.name}, reason: {reason}"
            )
            sys.exit(1)

    def recover(self) -> None:
        ok, reason = self.project.recover()
        if ok:
            console.print(f":clap: recover project {self.project.name}")
        else:
            console.print(
                f":fearful: failed to recover project {self.project.name}, reason: {reason}"
            )
            sys.exit(1)

    def info(self, fullname: bool = False) -> None:
        _r = self.project.info()
        _models = _r.pop("models", [])
        _datasets = _r.pop("datasets", [])

        def _show_objects(objects: t.List[t.Dict[str, t.Any]], typ: str) -> Tree:
            tree = Tree(f"[red]{typ}[/]")
            for _o in objects:
                otree = tree.add(f"{_o['name']}")
                for _v in _o["latest_versions"]:
                    _k = "name" if fullname else "short_name"
                    if typ == ProjectObjType.MODEL:
                        # TODO: add model version for every version
                        _size = _o["files"][0]["size"]
                    else:
                        _size = pretty_bytes(_v["meta"]["blobs_byte_size"])

                    otree.add(
                        f"[{_v['id']}][green]{_v[_k]}[/] :timer_clock: {_v['created_at']} :dizzy:{_size}"
                    )
            return tree

        console.print(Panel(Pretty(_r), title="Project Details", title_align="left"))
        if _models or _datasets:
            _block = self.comparison(
                _show_objects(_models, ProjectObjType.MODEL),
                _show_objects(_datasets, ProjectObjType.DATASET),
            )
            console.print(_block)


# TODO: add ProjectHTTPView for http request


class ProjectTermViewRich(ProjectTermView):
    @classmethod
    @BaseTermView._pager  # type: ignore
    @BaseTermView._header  # type: ignore
    def list(
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.Tuple[t.List[t.Any], t.Dict[str, t.Any]]:
        projects, pager = super().list(instance_uri, page, size)

        title = "Project List"
        custom_column: t.Dict[str, t.Callable[[t.Any], str]] = {
            "in_use": lambda x: ":backhand_index_pointing_right:" if x else "",
            "location": cls.place_holder_for_empty(),
            "owner": cls.place_holder_for_empty(),
        }
        custom_row: t.Callable[[t.Dict[str, t.Any]], t.Optional[t.Dict[str, str]]] = (
            lambda row: {"style": "magenta"} if row["in_use"] else None
        )
        cls.print_table(
            title, projects, custom_column=custom_column, custom_row=custom_row
        )
        return projects, pager

    @BaseTermView._header  # type: ignore
    def info(self, fullname: bool = False) -> None:
        _r = self.project.info()
        _models = _r.pop("models", [])
        _datasets = _r.pop("datasets", [])

        def _show_objects(objects: t.List[t.Dict[str, t.Any]], typ: str) -> Tree:
            tree = Tree(f"[red]{typ}[/]")
            for _o in objects:
                otree = tree.add(f"{_o['name']}")
                for _v in _o["latest_versions"]:
                    _k = "name" if fullname else "short_name"
                    if typ == ProjectObjType.MODEL:
                        # TODO: add model version for every version
                        _size = _o["files"][0]["size"]
                    else:
                        _size = pretty_bytes(_v["meta"]["blobs_byte_size"])

                    otree.add(
                        f"[{_v['id']}][green]{_v[_k]}[/] :timer_clock: {_v['created_at']} :dizzy:{_size}"
                    )
            return tree

        console.print(Panel(Pretty(_r), title="Project Details", title_align="left"))
        if _models or _datasets:
            _block = self.comparison(
                _show_objects(_models, ProjectObjType.MODEL),
                _show_objects(_datasets, ProjectObjType.DATASET),
            )
            console.print(_block)


class ProjectTermViewJson(ProjectTermView):
    @classmethod
    def list(  # type: ignore
        cls,
        instance_uri: str = "",
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> None:
        projects, pager = super().list(instance_uri, page, size)
        cls.pretty_json(projects)

    def info(self, fullname: bool = False) -> None:
        _r = self.project.info()
        self.pretty_json(_r)


def get_term_view(ctx_obj: t.Dict) -> t.Type[ProjectTermView]:
    return (
        ProjectTermViewJson if ctx_obj.get("output") == "json" else ProjectTermViewRich
    )



from __future__ import annotations

import typing as t
from pathlib import Path
from functools import wraps

from loguru import logger

from starwhale.consts import RunStatus, DecoratorInjectAttr
from starwhale.utils.load import load_module
from starwhale.utils.error import NoSupportError

from .context import Context

if t.TYPE_CHECKING:
    from .step import Step


class TaskResult:
    def __init__(self, id: int, status: str, exception: t.Optional[Exception] = None):
        self.id = id
        self.status = status
        self.exception = exception

    def __repr__(self) -> str:
        return f"id:{self.id}, status:{self.status}, exception:{self.exception}"


class TaskExecutor:
    def __init__(
        self,
        index: int,
        context: Context,
        workdir: Path,
        step: Step,
    ):
        self.index = index
        self.context = context
        self.workdir = workdir
        self.exception: t.Optional[Exception] = None
        self.step = step
        self.__status = RunStatus.INIT

    def __str__(self) -> str:
        return f"TaskExecutor[{self.index}]: step-{self.step}"

    def __repr__(self) -> str:
        return f"TaskExecutor[{self.index}]: step-{self.step}, workdir-{self.workdir}, context-{self.context}"

    @property
    def status(self) -> str:
        return self.__status

    def _get_internal_func_name(self, func_name: str) -> str:
        if func_name == "ppl":
            return "_starwhale_internal_run_ppl"
        elif func_name == "cmp":
            return "_starwhale_internal_run_cmp"
        else:
            raise RuntimeError(
                f"failed to map func name({func_name}) into PipelineHandler internal func name"
            )

    def _run_in_pipeline_handler_cls(
        self,
        func: t.Callable,
        func_name: str,
    ) -> None:
        from starwhale.api._impl.evaluation import PipelineHandler

        patch_func_map = {
            "ppl": lambda *args, **kwargs: ...,
            "cmp": lambda *args, **kwargs: ...,
        }

        if func_name not in patch_func_map:
            raise RuntimeError(
                f"func_name({func_name}) is not in patch func map:{patch_func_map.keys()}"
            )

        def _pop_self_args_func(f: t.Callable) -> t.Callable:
            @wraps(f)
            def _wrap(*args: t.Any, **kwargs: t.Any) -> t.Any:
                if len(args) >= 1 and isinstance(args[0], PipelineHandler):
                    args = args[1:]
                return f(*args, **kwargs)

            return _wrap

        patch_func_map[func_name] = _pop_self_args_func(func)

        cls_ = type(
            "GenericPipelineHandler",
            (PipelineHandler,),
            patch_func_map,
        )

        handler_func_name = self._get_internal_func_name(func_name)
        with cls_(*self.step.extra_args, **self.step.extra_kwargs) as instance:
            getattr(instance, handler_func_name)()

    def _do_execute(self) -> None:
        from starwhale.api._impl.evaluation import PipelineHandler

        module = load_module(self.step.module_name, self.workdir)
        cls_ = getattr(module, self.step.cls_name, None)

        if cls_ is None:
            func = getattr(module, self.step.func_name)
            if getattr(func, DecoratorInjectAttr.Evaluate, False):
                self._run_in_pipeline_handler_cls(func, "cmp")
            elif getattr(func, DecoratorInjectAttr.Predict, False):
                self._run_in_pipeline_handler_cls(func, "ppl")
            elif getattr(func, DecoratorInjectAttr.Step, False):
                func()
            else:
                raise NoSupportError(
                    f"func({self.step.module_name}.{self.step.func_name}) should use @step, @predict or @evaluate decorator"
                )
        else:
            # TODO: support user custom class and function with arguments
            if issubclass(cls_, PipelineHandler):
                func_name = self._get_internal_func_name(self.step.func_name)
            else:
                func_name = self.step.func_name

            if hasattr(cls_, "__enter__") and hasattr(cls_, "__exit__"):
                with cls_() as instance:
                    func = getattr(instance, func_name)
                    if getattr(func, DecoratorInjectAttr.Evaluate, False):
                        self._run_in_pipeline_handler_cls(func, "cmp")
                    elif getattr(func, DecoratorInjectAttr.Predict, False):
                        self._run_in_pipeline_handler_cls(func, "ppl")
                    else:
                        func()
            else:
                func = getattr(cls_(), func_name)
                if getattr(func, DecoratorInjectAttr.Evaluate, False):
                    self._run_in_pipeline_handler_cls(func, "cmp")
                elif getattr(func, DecoratorInjectAttr.Predict, False):
                    self._run_in_pipeline_handler_cls(func, "ppl")
                else:
                    func()

    def execute(self) -> TaskResult:
        logger.info(f"start to execute task with context({self.context}) ...")

        try:
            self.__status = RunStatus.RUNNING
            Context.set_runtime_context(self.context)
            self._do_execute()
        except Exception as e:
            logger.exception(e)
            self.exception = e
            self.__status = RunStatus.FAILED
        else:
            self.__status = RunStatus.SUCCESS
        finally:
            logger.info(
                f"finish {self.context}, status:{self.status}, error:{self.exception}"
            )
            return TaskResult(
                id=self.index, status=self.status, exception=self.exception
            )



from __future__ import annotations

import os
import copy
import json
import shutil
import typing as t
import tarfile
from abc import ABCMeta
from pathlib import Path
from collections import defaultdict

import yaml
from fs import open_fs
from loguru import logger
from fs.walk import Walker

from starwhale import PipelineHandler
from starwhale.utils import console, now_str, load_yaml, gen_uniq_version
from starwhale.consts import (
    FileDesc,
    FileFlag,
    FileNode,
    RunStatus,
    CREATED_AT_KEY,
    SWMP_SRC_FNAME,
    DefaultYAMLName,
    SW_AUTO_DIRNAME,
    DEFAULT_PAGE_IDX,
    DEFAULT_PAGE_SIZE,
    SW_IGNORE_FILE_NAME,
    DEFAULT_MANIFEST_NAME,
    DEFAULT_JOBS_FILE_NAME,
    DEFAULT_FINETUNE_JOB_NAME,
    SW_EVALUATION_EXAMPLE_DIR,
    DEFAULT_EVALUATION_JOB_NAME,
    DEFAULT_STARWHALE_API_VERSION,
    EVALUATION_SVC_META_FILE_NAME,
    EVALUATION_PANEL_LAYOUT_JSON_FILE_NAME,
    EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME,
    DEFAULT_FILE_SIZE_THRESHOLD_TO_TAR_IN_MODEL,
)
from starwhale.base.tag import StandaloneTag
from starwhale.base.uri import URI
from starwhale.utils.fs import (
    move_dir,
    copy_file,
    file_stat,
    ensure_dir,
    ensure_file,
    blake2b_file,
)
from starwhale.base.type import URIType, BundleType, InstanceType
from starwhale.base.cloud import CloudRequestMixed, CloudBundleModelMixin
from starwhale.base.mixin import ASDictMixin
from starwhale.utils.http import ignore_error
from starwhale.utils.load import load_module
from starwhale.api.service import Service
from starwhale.base.bundle import BaseBundle, LocalStorageBundleMixin
from starwhale.utils.error import NoSupportError, FileFormatError
from starwhale.api._impl.job import generate_jobs_yaml
from starwhale.core.job.step import Step
from starwhale.utils.progress import run_with_progress_bar
from starwhale.base.blob.store import LocalFileStore
from starwhale.core.eval.store import EvaluationStorage
from starwhale.core.model.copy import ModelCopy
from starwhale.core.job.context import Context
from starwhale.core.model.store import ModelStorage
from starwhale.api._impl.service import Hijack
from starwhale.core.job.scheduler import Scheduler


class ModelRunConfig(ASDictMixin):
    # TODO: use attr to tune class
    def __init__(
        self,
        handler: str,
        runtime: str = "",
        envs: t.Union[t.List[str], None] = None,
        **kw: t.Any,
    ):
        self.handler = handler.strip()
        self.runtime = runtime.strip()
        self.envs = envs or []
        self.kw = kw

        self._do_validate()

    def _do_validate(self) -> None:
        if not self.handler:
            raise FileFormatError("need ppl field")

    def __str__(self) -> str:
        return f"Model Run Config: ppl -> {self.handler}"

    def __repr__(self) -> str:
        return f"Model Run Config: ppl -> {self.handler}, runtime -> {self.runtime}"


class ModelConfig(ASDictMixin):
    # TODO: use attr to tune class
    def __init__(
        self,
        name: str,
        run: t.Dict[str, t.Any],
        desc: str = "",
        tag: t.Optional[t.List[str]] = None,
        version: str = DEFAULT_STARWHALE_API_VERSION,
        **kw: t.Any,
    ):
        # TODO: format model name
        self.name = name
        # TODO: support artifacts: local or remote
        self.run = ModelRunConfig(**run)
        self.desc = desc
        self.tag = tag or []
        self.version = version
        self.kw = kw

    @classmethod
    def create_by_yaml(cls, path: Path) -> ModelConfig:
        c = load_yaml(path)
        return cls(**c)

    def __str__(self) -> str:
        return f"Model Config: {self.name}"

    def __repr__(self) -> str:
        return f"Model Config: name -> {self.name}"


class Model(BaseBundle, metaclass=ABCMeta):
    def __str__(self) -> str:
        return f"Starwhale Model: {self.uri}"

    def eval(self) -> None:
        pass

    def ppl(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.MODEL, **kw: t.Any
    ) -> None:
        pass

    def cmp(
        self, workdir: Path, yaml_name: str = DefaultYAMLName.MODEL, **kw: t.Any
    ) -> None:
        pass

    @classmethod
    def get_model(cls, uri: URI) -> Model:
        _cls = cls._get_cls(uri)
        return _cls(uri)

    @classmethod
    def _get_cls(cls, uri: URI) -> t.Union[t.Type[StandaloneModel], t.Type[CloudModel]]:  # type: ignore
        if uri.instance_type == InstanceType.STANDALONE:
            return StandaloneModel
        elif uri.instance_type == InstanceType.CLOUD:
            return CloudModel
        else:
            raise NoSupportError(f"model uri:{uri}")

    @classmethod
    def copy(
        cls,
        src_uri: str,
        dest_uri: str,
        force: bool = False,
        dest_local_project_uri: str = "",
    ) -> None:
        bc = ModelCopy(
            src_uri,
            dest_uri,
            URIType.MODEL,
            force,
            dest_local_project_uri=dest_local_project_uri,
        )
        bc.do()

    def diff(self, compare_uri: URI) -> t.Dict[str, t.Any]:
        raise NotImplementedError


def resource_to_file_node(
    files: t.List[t.Dict[str, t.Any]], parent_path: Path
) -> t.Dict[str, FileNode]:
    return {
        _f["path"]: FileNode(
            path=parent_path / _f["path"],
            name=_f.get("name") or os.path.basename(_f["path"]),
            size=_f.get("size") or file_stat(parent_path / _f["path"]).st_size,
            signature=_f.get("signature") or blake2b_file(parent_path / _f["path"]),
            file_desc=FileDesc[_f["desc"]],
        )
        for _f in files
    }


class StandaloneModel(Model, LocalStorageBundleMixin):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.STANDALONE
        self.store = ModelStorage(uri)
        self.tag = StandaloneTag(uri)
        self._manifest: t.Dict[str, t.Any] = {}  # TODO: use manifest class
        self.models: t.List[t.Dict[str, t.Any]] = []
        self.sources: t.List[t.Dict[str, t.Any]] = []
        self.yaml_name = DefaultYAMLName.MODEL
        self._version = uri.object.version
        self._object_store = LocalFileStore()

    def list_tags(self) -> t.List[str]:
        return self.tag.list()

    def add_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.add(tags, ignore_errors)

    def remove_tags(self, tags: t.List[str], ignore_errors: bool = False) -> None:
        self.tag.remove(tags, ignore_errors)

    def _gen_model_serving(self, ppl: str, workdir: Path) -> None:
        rc_dir = str(
            self.store.hidden_sw_dir.relative_to(self.store.snapshot_workdir)
            / SW_EVALUATION_EXAMPLE_DIR
        )
        # render spec
        svc = self._get_service(ppl, workdir, hijack=Hijack(True, rc_dir))
        file = self.store.hidden_sw_dir / EVALUATION_SVC_META_FILE_NAME
        ensure_file(file, json.dumps(svc.get_spec(), indent=4), parents=True)

        if len(svc.example_resources) == 0:
            return

        # check duplicate file names, do not support using examples with same name in different dir
        names = set([os.path.basename(i) for i in svc.example_resources])
        if len(names) != len(svc.example_resources):
            raise NoSupportError("duplicate file names in examples")

        # copy example resources for online evaluation in server instance
        dst = self.store.hidden_sw_dir / SW_EVALUATION_EXAMPLE_DIR
        ensure_dir(dst)
        for f in svc.example_resources:
            shutil.copy2(f, dst)

    def _render_eval_layout(self, workdir: Path) -> None:
        # render eval layout
        eval_layout = workdir / SW_AUTO_DIRNAME / EVALUATION_PANEL_LAYOUT_YAML_FILE_NAME
        if eval_layout.exists():
            content = load_yaml(eval_layout)
            dst = self.store.hidden_sw_dir / EVALUATION_PANEL_LAYOUT_JSON_FILE_NAME
            ensure_file(dst, json.dumps(content), parents=True)

    @staticmethod
    def _get_service(
        module: str, pkg: Path, hijack: t.Optional[Hijack] = None
    ) -> Service:
        module, _, attr = module.partition(":")
        m = load_module(module, pkg)
        apis = dict()
        ins: t.Any = None
        svc: t.Optional[Service] = None

        # TODO: refine this ugly ad hoc
        Context.set_runtime_context(
            Context(pkg, version="-1", project="tmp-project-for-build")
        )

        # TODO: check duplication
        for k, v in m.__dict__.items():
            if isinstance(v, Service):
                apis.update(v.apis)
                # use Service in module
                svc = v
        if attr:
            cls = getattr(m, attr)
            ins = cls()
            if isinstance(ins, PipelineHandler):
                apis.update(ins.svc.apis)

        from starwhale.api._impl.service import internal_api_list

        apis.update(internal_api_list())

        # check if we need to instance the model when using custom handler
        if ins is None:
            for i in apis.values():
                fn = i.func.__qualname__
                # TODO: support deep path classes (also modules)
                if "." in fn:
                    ins = getattr(m, fn.split(".", 1)[0])()
                    break

        if svc is None:
            svc = Service()
        for api in apis.values():
            svc.add_api_instance(api)
        svc.api_instance = ins
        svc.hijack = hijack
        return svc

    @classmethod
    def eval_user_handler(
        cls,
        project: str,
        version: str,
        workdir: Path,
        dataset_uris: t.List[str],
        model_yaml_name: str = DefaultYAMLName.MODEL,
        job_name: str = DEFAULT_EVALUATION_JOB_NAME,
        step_name: str = "",
        task_index: t.Optional[int] = None,
        task_num: int = 0,
        base_info: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> None:
        base_info = base_info or {}
        _manifest: t.Dict[str, t.Any] = {
            CREATED_AT_KEY: now_str(),
            "status": RunStatus.START,
            "step": step_name,
            "task_index": task_index,
            "task_num": task_num,
        }
        _model_config = cls.load_model_config(workdir / model_yaml_name, workdir)

        if not version:
            version = gen_uniq_version()

        _project_uri = URI(project, expected_type=URIType.PROJECT)
        _run_dir = EvaluationStorage.local_run_dir(_project_uri.project, version)
        ensure_dir(_run_dir)

        yaml_path = workdir / SW_AUTO_DIRNAME / DEFAULT_JOBS_FILE_NAME

        if not yaml_path.exists():
            # do not auto generate eval_job.yaml in the user workdir
            generate_jobs_yaml(
                run_handler=_model_config.run.handler,
                workdir=workdir,
                yaml_path=yaml_path,
            )

        logger.debug(f"parse job from yaml:{yaml_path}")

        console.print(
            f":hourglass_not_done: start to run evaluation[{job_name}:{version}]..."
        )
        _scheduler = Scheduler(
            project=_project_uri.project,
            version=version,
            workdir=workdir,
            dataset_uris=dataset_uris,
            steps=Step.get_steps_from_yaml(job_name, yaml_path),
        )
        _status = RunStatus.START
        try:
            _results = _scheduler.run(
                step_name=step_name, task_index=task_index, task_num=task_num
            )
            _status = RunStatus.SUCCESS

            exceptions: t.List[Exception] = []
            for _r in _results:
                for _tr in _r.task_results:
                    if _tr.exception:
                        exceptions.append(_tr.exception)
            if exceptions:
                raise Exception(*exceptions)

            logger.debug(
                f"-->[Finished] evaluation[{job_name}:{version}] execute finished, results info:{_results}"
            )
        except Exception as e:
            _status = RunStatus.FAILED
            _manifest["error_message"] = str(e)
            logger.error(
                f"-->[Failed] evaluation[{job_name}:{version}] execute failed, error info:{e}"
            )
            raise
        finally:
            _manifest.update(
                {
                    **dict(
                        version=version,
                        project=_project_uri.project,
                        model_dir=str(workdir),
                        datasets=list(dataset_uris),
                        status=_status,
                        finished_at=now_str(),
                    ),
                    **base_info,
                }
            )
            _f = _run_dir / DEFAULT_MANIFEST_NAME
            ensure_file(_f, yaml.safe_dump(_manifest, default_flow_style=False))

            console.print(
                f":{100 if _status == RunStatus.SUCCESS else 'broken_heart'}: finish run, {_status}!"
            )

    @classmethod
    def fine_tune(
        cls,
        project: str,
        workdir: Path,
        dataset_uris: t.List[str],
        model_yaml_name: str = DefaultYAMLName.MODEL,
        job_name: str = DEFAULT_FINETUNE_JOB_NAME,
        step_name: str = "fine_tune",
        task_index: int = 0,
        task_num: int = 1,
    ) -> None:
        _model_config = cls.load_model_config(workdir / model_yaml_name, workdir)

        _project_uri = URI(project, expected_type=URIType.PROJECT)

        yaml_path = workdir / SW_AUTO_DIRNAME / DEFAULT_JOBS_FILE_NAME

        if not yaml_path.exists():
            # do not auto generate eval_job.yaml in the user workdir
            generate_jobs_yaml(
                run_handler=_model_config.run.handler,
                workdir=workdir,
                yaml_path=yaml_path,
            )

        logger.debug(f"parse fine-tune job from yaml:{yaml_path}")
        version = gen_uniq_version()
        console.print(
            f":hourglass_not_done: start to run fine-tune[{job_name}:{version}]..."
        )
        _scheduler = Scheduler(
            project=_project_uri.project,
            version=version,
            workdir=workdir,
            dataset_uris=dataset_uris,
            steps=Step.get_steps_from_yaml(job_name, yaml_path),
        )
        _status = RunStatus.START
        try:
            _results = _scheduler.run(
                step_name=step_name, task_index=task_index, task_num=task_num
            )
            _status = RunStatus.SUCCESS

            exceptions: t.List[Exception] = []
            for _r in _results:
                for _tr in _r.task_results:
                    if _tr.exception:
                        exceptions.append(_tr.exception)
            if exceptions:
                raise Exception(*exceptions)

            logger.debug(
                f"-->[Finished] fine-tune[{job_name}:{version}] execute finished, results info:{_results}"
            )
        except Exception as e:
            _status = RunStatus.FAILED
            logger.error(
                f"-->[Failed] fine-tune[{job_name}:{version}] execute failed, error info:{e}"
            )
            raise
        finally:
            console.print(
                f":{100 if _status == RunStatus.SUCCESS else 'broken_heart'}: finish run, {_status}!"
            )

    def diff(self, compare_uri: URI) -> t.Dict[str, t.Any]:
        """
        - added: a node that exists in compare but not in base
        - deleted: a node that not exists in compare but in base
        - updated: a node that exists in both of base and compare but signature is different
        - unchanged: a node that exists in both of base and compare, and signature is same
        :param compare_uri:
        :return: diff info
        """
        # TODO use remote get model info for cloud
        if compare_uri.instance_type != InstanceType.STANDALONE:
            raise NoSupportError(
                f"only support standalone uri, but compare_uri({compare_uri}) is for cloud instance"
            )
        if self.uri.object.name != compare_uri.object.name:
            raise NoSupportError(
                f"only support two versions diff in one model, base model:{self.uri}, compare model:{compare_uri}"
            )
        _compare_model = StandaloneModel(compare_uri)
        base_file_maps = resource_to_file_node(
            files=self.store.manifest["resources"],
            parent_path=self.store.snapshot_workdir,
        )
        compare_file_maps = resource_to_file_node(
            files=_compare_model.store.manifest["resources"],
            parent_path=_compare_model.store.snapshot_workdir,
        )
        all_paths = {
            p for m in (base_file_maps, compare_file_maps) for p, _ in m.items()
        }

        for _p in all_paths:
            if _p in base_file_maps and _p in compare_file_maps:
                if base_file_maps[_p].signature == compare_file_maps[_p].signature:
                    compare_file_maps[_p].flag = FileFlag.UNCHANGED
                else:
                    compare_file_maps[_p].flag = FileFlag.UPDATED
            if _p in base_file_maps and _p not in compare_file_maps:
                del_file = copy.copy(base_file_maps[_p])
                del_file.flag = FileFlag.DELETED
                compare_file_maps[_p] = del_file
            if _p not in base_file_maps and _p in compare_file_maps:
                compare_file_maps[_p].flag = FileFlag.ADDED

        return {
            "all_paths": all_paths,
            "base_version": base_file_maps,
            "compare_version": compare_file_maps,
        }

    def info(self) -> t.Dict[str, t.Any]:
        _manifest = self._get_bundle_info()
        _store = self.store
        _om = {}
        yaml_path = _store.hidden_sw_dir / DEFAULT_JOBS_FILE_NAME
        if _store.snapshot_workdir.exists():
            if yaml_path.exists():
                _om = load_yaml(yaml_path)
            else:
                logger.warning("step_spec not found in model snapshot_workdir")
        else:
            logger.warning("step_spec not found in model")
        if _om:
            _manifest["step_spec"] = _om
        return _manifest

    def history(
        self,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
    ) -> t.List[t.Dict[str, t.Any]]:
        _r = []
        for _bf in self.store.iter_bundle_history():
            _manifest_path = _bf.path / DEFAULT_MANIFEST_NAME
            if not _manifest_path.exists():
                continue

            _manifest = load_yaml(_manifest_path)

            _r.append(
                dict(
                    name=self.name,
                    version=_bf.version,
                    path=str(_bf.path.resolve()),
                    tags=_bf.tags,
                    created_at=_manifest[CREATED_AT_KEY],
                    size=_bf.path.stat().st_size,
                )
            )
        return _r

    def remove(self, force: bool = False) -> t.Tuple[bool, str]:
        return self._do_remove(force)

    def recover(self, force: bool = False) -> t.Tuple[bool, str]:
        # TODO: support short version to recover, today only support full-version
        dest_path = (
            self.store.bundle_dir / f"{self.uri.object.version}{BundleType.MODEL}"
        )
        _ok, _reason = move_dir(self.store.recover_loc, dest_path, force)
        _ok2, _reason2 = True, ""
        if self.store.recover_snapshot_workdir.exists():
            _ok2, _reason2 = move_dir(
                self.store.recover_snapshot_workdir, self.store.snapshot_workdir, force
            )
        return _ok and _ok2, _reason + _reason2

    @classmethod
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filters: t.Optional[t.Union[t.Dict[str, t.Any], t.List[str]]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filters = filters or {}
        rs = defaultdict(list)
        for _bf in ModelStorage.iter_all_bundles(
            project_uri,
            bundle_type=BundleType.MODEL,
            uri_type=URIType.MODEL,
        ):
            if not cls.do_bundle_filter(_bf, filters):
                continue

            if _bf.path.is_file():
                # for origin swmp(tar)
                _manifest = ModelStorage.get_manifest_by_path(
                    _bf.path, BundleType.MODEL, URIType.MODEL
                )
            elif (_bf.path / DEFAULT_MANIFEST_NAME).exists():
                _manifest = load_yaml(_bf.path / DEFAULT_MANIFEST_NAME)
            else:
                continue
            rs[_bf.name].append(
                {
                    "name": _bf.name,
                    "version": _bf.version,
                    "path": str(_bf.path.absolute()),
                    "size": _bf.path.stat().st_size,
                    "is_removed": _bf.is_removed,
                    CREATED_AT_KEY: _manifest[CREATED_AT_KEY],
                    "tags": _bf.tags,
                }
            )
        return rs, {}

    def buildImpl(self, workdir: Path, **kw: t.Any) -> None:  # type: ignore[override]
        yaml_path = Path(kw.get("yaml_path", workdir / DefaultYAMLName.MODEL))
        _model_config = self.load_model_config(yaml_path, workdir)

        logger.debug(f"build workdir:{workdir}")

        operations = [
            (self._gen_version, 5, "gen version"),
            (self._prepare_snapshot, 5, "prepare snapshot"),
            (
                self._copy_src,
                15,
                "copy src",
                dict(workdir=workdir, yaml_path=yaml_path),
            ),
            (
                generate_jobs_yaml,
                5,
                "generate jobs yaml",
                dict(
                    run_handler=_model_config.run.handler,
                    workdir=workdir,
                    yaml_path=self.store.src_dir
                    / SW_AUTO_DIRNAME
                    / DEFAULT_JOBS_FILE_NAME,
                ),
            ),
            (
                self._gen_model_serving,
                10,
                "generate model serving",
                dict(ppl=_model_config.run.handler, workdir=workdir),
            ),
            (
                self._render_eval_layout,
                1,
                "render eval layout",
                dict(workdir=workdir),
            ),
            (
                self._make_meta_tar,
                20,
                "build model bundle",
            ),
            (
                self._render_manifest,
                5,
                "render manifest",
            ),
            (self._make_auto_tags, 5, "make auto tags"),
        ]
        run_with_progress_bar("model bundle building...", operations)

    def _make_meta_tar(
        self, size_th_to_tar: int = DEFAULT_FILE_SIZE_THRESHOLD_TO_TAR_IN_MODEL
    ) -> None:
        w = Walker()
        src_fs = open_fs(str(self.store.src_dir.resolve()))
        with tarfile.open(self.store.snapshot_workdir / SWMP_SRC_FNAME, "w:") as tar:
            for f in w.files(src_fs):
                sub_path = f[1:]
                size = file_stat(self.store.src_dir / sub_path).st_size
                separate = size > size_th_to_tar
                file_info = {
                    "name": os.path.basename(f),
                    "path": f"{self.store.src_dir_name}/{sub_path}",
                    "signature": blake2b_file(self.store.src_dir / sub_path),
                    "duplicate_check": separate,
                    "desc": FileDesc.SRC.name if not separate else FileDesc.MODEL.name,
                    "size": size,
                }
                if separate:
                    self.models.append(file_info)
                else:
                    tar.add(str(self.store.src_dir / sub_path), arcname=sub_path)
                    self.sources.append(file_info)

    def _render_manifest(self) -> None:
        self._manifest["resources"] = self.models + self.sources
        super()._render_manifest()

    @classmethod
    def load_model_config(cls, yaml_path: Path, workdir: Path) -> ModelConfig:
        cls._do_validate_yaml(yaml_path)
        _config = ModelConfig.create_by_yaml(yaml_path)

        # TODO: add more model.yaml section validation
        # TODO: add 'swcli model check' cmd

        cls._load_config_envs(_config)
        return _config

    def _prepare_snapshot(self) -> None:
        logger.info("[step:prepare-snapshot]prepare model snapshot dirs...")

        ensure_dir(self.store.snapshot_workdir)
        ensure_dir(self.store.src_dir)

        # TODO: cleanup garbage dir
        # TODO: add lock/flag file for gc

        console.print(
            f":file_folder: workdir: [underline]{self.store.snapshot_workdir}[/]"
        )

    def _copy_src(self, workdir: Path, yaml_path: Path) -> None:
        logger.info(
            f"[step:copy]start to copy src {workdir} -> {self.store.src_dir} ..."
        )
        console.print(":thumbs_up: try to copy source code files...")

        excludes = None
        ignore = workdir / SW_IGNORE_FILE_NAME
        if ignore.exists():
            with open(ignore, "r") as f:
                excludes = [line.strip() for line in f.readlines()]
        self._object_store.copy_dir(
            str(workdir.resolve()), str(self.store.src_dir.resolve()), excludes=excludes
        )

        # TODO: use an internal protected model yaml path
        # make sure model.yaml exists, prevent using the wrong config when running in the cloud
        target_yaml_path = self.store.src_dir / DefaultYAMLName.MODEL
        copy_file(yaml_path, target_yaml_path)

        logger.info("[step:copy]finish copy files")

    @classmethod
    def _load_config_envs(cls, _config: ModelConfig) -> None:
        for _env in _config.run.envs:
            _env = _env.strip()
            if not _env:
                continue
            _t = _env.split("=", 1)
            _k, _v = _t[0], "".join(_t[1:])

            if _k not in os.environ:
                os.environ[_k] = _v

    @classmethod
    def serve(
        cls,
        model_yaml: str,
        workdir: Path,
        host: str,
        port: int,
    ) -> None:
        _model_config = cls.load_model_config(workdir / model_yaml, workdir)
        svc = cls._get_service(_model_config.run.handler, workdir)
        svc.serve(host, port, _model_config.name)


class CloudModel(CloudBundleModelMixin, Model):
    def __init__(self, uri: URI) -> None:
        super().__init__(uri)
        self.typ = InstanceType.CLOUD

    @classmethod
    @ignore_error(({}, {}))
    def list(
        cls,
        project_uri: URI,
        page: int = DEFAULT_PAGE_IDX,
        size: int = DEFAULT_PAGE_SIZE,
        filter_dict: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Tuple[t.Dict[str, t.Any], t.Dict[str, t.Any]]:
        filter_dict = filter_dict or {}
        crm = CloudRequestMixed()
        return crm._fetch_bundle_all_list(
            project_uri, URIType.MODEL, page, size, filter_dict
        )

    def build(self, *args: t.Any, **kwargs: t.Any) -> None:
        raise NoSupportError("no support build model in the cloud instance")

    def diff(self, compare_uri: URI) -> t.Dict[str, t.Any]:
        raise NoSupportError("no support model diff in the cloud instance")



import os
import sys
import typing as t
import getpass
import subprocess
from shutil import which
from pathlib import Path

import yaml

from starwhale.utils import load_yaml
from starwhale.consts import (
    UserRoleType,
    SW_CLI_CONFIG,
    DEFAULT_PROJECT,
    DEFAULT_INSTANCE,
    ENV_SW_CLI_CONFIG,
    DATA_STORE_DIRNAME,
    STANDALONE_INSTANCE,
    ENV_SW_LOCAL_STORAGE,
    LOCAL_CONFIG_VERSION,
    OBJECT_STORE_DIRNAME,
    DEFAULT_SW_LOCAL_STORAGE,
)
from starwhale.consts.env import SWEnv
from starwhale.utils.error import NotFoundError, NoSupportError

from . import console, now_str, fmt_http_server
from .fs import ensure_dir, ensure_file

_config: t.Dict[str, t.Any] = {}
_CURRENT_SHELL_USERNAME = getpass.getuser()


def load_swcli_config() -> t.Dict[str, t.Any]:
    global _config

    if _config:
        ensure_dir(Path(_config["storage"]["root"]) / DEFAULT_PROJECT, recursion=True)
        return _config

    # TODO: add set_global_env func in cli startup
    fpath = get_swcli_config_path()

    if not os.path.exists(fpath):
        _config = render_default_swcli_config(fpath)
    else:
        _config = load_yaml(fpath)

        env = os.environ.get(ENV_SW_LOCAL_STORAGE)
        if env:
            _config["storage"]["root"] = env

        _version = _config.get("version")
        if _version != LOCAL_CONFIG_VERSION:
            console.print(
                f":cherries: {fpath} use unexpected version({_version}), swcli only support {LOCAL_CONFIG_VERSION} version."
            )
            console.print(
                f":carrot: {fpath} will be upgraded to {LOCAL_CONFIG_VERSION} automatically."
            )
            _config = render_default_swcli_config(fpath)

    ensure_dir(Path(_config["storage"]["root"]) / DEFAULT_PROJECT, recursion=True)
    return _config


def render_default_swcli_config(fpath: str) -> t.Dict[str, t.Any]:
    from starwhale.base.type import InstanceType

    env_root = os.environ.get(ENV_SW_LOCAL_STORAGE)
    c = dict(
        version=LOCAL_CONFIG_VERSION,
        instances={
            STANDALONE_INSTANCE: dict(
                uri=DEFAULT_INSTANCE,
                user_name=_CURRENT_SHELL_USERNAME,
                current_project=DEFAULT_PROJECT,
                type=InstanceType.STANDALONE,
                updated_at=now_str(),
            )
        },
        current_instance=DEFAULT_INSTANCE,
        storage=dict(root=env_root or str(DEFAULT_SW_LOCAL_STORAGE.resolve())),
    )
    render_swcli_config(c, fpath)
    return c


def update_swcli_config(**kw: t.Any) -> None:
    c = load_swcli_config()
    # TODO: tune update config
    # TODO: add deepcopy for dict?
    c.update(kw)
    render_swcli_config(c)


def get_swcli_config_path() -> str:
    return os.environ.get(ENV_SW_CLI_CONFIG, "") or str(SW_CLI_CONFIG)


def render_swcli_config(c: t.Dict[str, t.Any], path: str = "") -> None:
    fpath = path or get_swcli_config_path()
    ensure_dir(os.path.dirname(fpath), recursion=True)
    ensure_file(fpath, yaml.safe_dump(c, default_flow_style=False), mode=0o600)


# TODO: abstract better common base or mixed class
class SWCliConfigMixed:
    def __init__(self, swcli_config: t.Optional[t.Dict[str, t.Any]] = None) -> None:
        self._config = swcli_config or load_swcli_config()

    @property
    def rootdir(self) -> Path:
        return Path(self._config["storage"]["root"])

    @property
    def datastore_dir(self) -> Path:
        return self.rootdir / DATA_STORE_DIRNAME

    @property
    def object_store_dir(self) -> Path:
        return self.rootdir / OBJECT_STORE_DIRNAME

    @property
    def sw_remote_addr(self) -> str:
        addr: str = self._current_instance_obj.get("uri", "")
        if addr == "local":
            return addr
        else:
            return fmt_http_server(addr)

    @property
    def user_name(self) -> str:
        return self._current_instance_obj.get("user_name", "")

    @property
    def _sw_token(self) -> str:
        return self._current_instance_obj.get("sw_token", "") or os.environ.get(
            SWEnv.instance_token, ""
        )

    @property
    def _current_instance_obj(self) -> t.Dict[str, t.Any]:
        return self._config.get("instances", {}).get(self.current_instance) or {}

    @property
    def user_role(self) -> str:
        return self._current_instance_obj.get("user_role", "")

    @property
    def current_instance(self) -> str:
        return str(self._config["current_instance"])

    @property
    def link_auths(self) -> t.Any:
        return self._config.get("link_auths")

    def get_sw_instance_config(self, instance: str) -> t.Dict[str, t.Any]:
        instance = self._get_instance_alias(instance)
        return self._config["instances"].get(instance) or {}

    def get_sw_token(self, instance: str) -> str:
        return self.get_sw_instance_config(instance).get("sw_token", "")

    def _get_instance_alias(self, instance: str) -> str:
        if not instance:
            return self.current_instance

        if instance not in self._config["instances"]:
            _count = 0
            _alias = None
            for k, v in self._config["instances"].items():
                if v["uri"] == instance:
                    _alias = str(k)
                    _count += 1
            if _count > 1:
                raise RuntimeError(
                    f"instance uri:{instance} has multi items!! please use alias."
                )
            if _alias:
                return _alias

        return instance

    @property
    def current_project(self) -> str:
        return self._current_instance_obj.get("current_project", "")

    def select_current_default(self, instance: str, project: str = "") -> None:
        instance = self._get_instance_alias(instance)

        if instance not in self._config["instances"]:
            raise NotFoundError(f"need to login instance {instance}")

        self._config["current_instance"] = instance
        if project:
            if (
                instance == STANDALONE_INSTANCE
                and not (self.rootdir / project).exists()
            ):
                raise NotFoundError(f"need to create project {project}")
            # TODO: check cloud project existence
            self._config["instances"][instance]["current_project"] = project

        update_swcli_config(**self._config)

    def delete_instance(self, uri: str) -> None:
        if uri == STANDALONE_INSTANCE:
            return

        _insts = self._config["instances"]
        _alias = uri
        if uri in _insts:
            _insts.pop(uri)
        else:
            for k, v in _insts.items():
                if v.get("uri") == uri:
                    _insts.pop(k)
                    _alias = uri

        if _alias == self._config["current_instance"]:
            self._config["current_instance"] = DEFAULT_INSTANCE
        update_swcli_config(**self._config)

    def update_instance(
        self,
        uri: str,
        user_name: str = _CURRENT_SHELL_USERNAME,
        user_role: str = UserRoleType.NORMAL,
        sw_token: str = "",
        alias: str = "",
    ) -> None:
        from starwhale.base.type import InstanceType

        # TODO: abstrace instance class
        uri = uri.strip()
        if not uri.startswith(("http://", "https://")):
            uri = f"http://{uri}"

        alias = alias or uri
        if alias == STANDALONE_INSTANCE:
            console.print(f":person_running: skip {STANDALONE_INSTANCE} update")
            return

        # TODO: add more instance list and search
        _instances: t.Dict[str, t.Dict[str, str]] = self._config["instances"]
        if alias not in _instances:
            _instances[alias] = {}

        _instances[alias].update(
            uri=uri,
            user_name=user_name,
            user_role=user_role,
            sw_token=sw_token,
            type=InstanceType.CLOUD,
            updated_at=now_str(),
        )

        update_swcli_config(**self._config)


def edit_from_shell() -> None:
    _editor = os.environ.get("EDITOR") or os.environ.get("VISUAL") or "vi"
    if which(_editor.split()[0]) is None:
        raise NoSupportError(
            f"no found {_editor} bin in {sys.platform}. Please configure one using EDITOR or VISUAL environment variable"
        )
    subprocess.call(f"{_editor} {get_swcli_config_path()}", shell=True)



from typing import List, Optional


class NoMatchException(Exception):
    def __init__(self, item: str, found: Optional[List[str]] = None) -> None:
        message = f"Can not find the exact match item {item}, found: {found}"
        super().__init__(message)


class UriTooShortException(Exception):
    def __init__(self, expect: int, get: int, msg: str = "") -> None:
        super().__init__(
            ", ".join([f"URI too short, expect {expect} parts, get {get}", msg])
        )



import os

try:
    from importlib.metadata import version
except ImportError:  # pragma: no cover
    from importlib_metadata import version  # pragma: no cover


STARWHALE_VERSION: str = version("starwhale")  # type: ignore
os.environ["SW_VERSION"] = STARWHALE_VERSION
